# ===========================================================================================================
# ------------------------------ GENERIC TECH DRC-malformed RULE DECK ---------------------------------------
# ===========================================================================================================

"""
Run GENERIC TECH DRC-malformed runset.

Usage:
    run_drc.py (--help| -h)
    run_drc.py (--path=<file_path>) [--verbose] [--mp=<num_cores>] [--run_dir=<run_dir_path>] [--topcell=<topcell_name>] [--thr=<thr>] [--run_mode=<run_mode>] [--offgrid]

Options:
    --help -h                           Print this help message.
    --path=<file_path>                  The input GDS file path.
    --topcell=<topcell_name>            Topcell name to use.
    --mp=<num_cores>                    Run the rule deck in parts in parallel to speed up the run. [default: 1]
    --run_dir=<run_dir_path>            Run directory to save all the results [default: pwd]
    --thr=<thr>                         The number of threads used in run.
    --run_mode=<run_mode>               Select klayout mode Allowed modes (flat , deep, tiling). [default: deep]
    --verbose                           Detailed rule execution log for debugging.
"""

import logging
import os
import xml.etree.ElementTree as ET
from datetime import datetime
from subprocess import check_call

import klayout.db
from docopt import docopt


def get_rules_with_violations(results_database):
    """
    This function will find all the rules that has violated in a database.

    Parameters
    ----------
    results_database : string or Path object
        Path string to the results file

    Returns
    -------
    set
        A set that contains all rules in the database with violations
    """

    mytree = ET.parse(results_database)
    myroot = mytree.getroot()

    all_violating_rules = set()

    for z in myroot[7]:  # myroot[7] : List rules with viloations
        all_violating_rules.add(f"{z[1].text}".replace("'", ""))

    return all_violating_rules


def check_drc_results(results_db_files: list) -> None:
    """
    check_drc_results Checks the results db generated from run and report at the end if the DRC run failed or passed.
    This function will exit with 1 if there are violations.

    Parameters
    ----------
    results_db_files : list
        A list of strings that represent paths to results databases of all the DRC runs.
    """

    if len(results_db_files) < 1:
        logging.error("Klayout did not generate any rdb results. Please check run logs")
        exit(1)

    full_violating_rules = set()

    for f in results_db_files:
        violating_rules = get_rules_with_violations(f)
        full_violating_rules.update(violating_rules)

    if len(full_violating_rules) > 0:
        logging.error("Klayout DRC run is not clean.")
        logging.error(f"Violated rules are : {str(full_violating_rules)}\n")
        exit(1)
    else:
        logging.info("Klayout DRC run is clean. GDS has no DRC violations.")


def get_top_cell_names(gds_path):
    """
    get_top_cell_names get the top cell names from the GDS file.

    Parameters
    ----------
    gds_path : string
        Path to the target GDS file.

    Returns
    -------
    List of string
        Names of the top cell in the layout.
    """
    layout = klayout.db.Layout()
    layout.read(gds_path)
    top_cells = [t.name for t in layout.top_cells()]

    return top_cells


def get_run_top_cell_name(arguments, layout_path):
    """
    get_run_top_cell_name Get the top cell name to use for running. If it's provided by the user, we use the user input.
    If not, we get it from the GDS file.

    Parameters
    ----------
    arguments : dict
        Dictionary that holds the user inputs for the script generated by docopt.
    layout_path : string
        Path to the target layout.

    Returns
    -------
    string
        Name of the topcell to use in run.

    """

    if arguments["--topcell"]:
        topcell = arguments["--topcell"]
    else:
        layout_topcells = get_top_cell_names(layout_path)
        if len(layout_topcells) > 1:
            logging.error(
                "## Layout has multiple topcells. Please determine which topcell you want to run on."
            )
            exit(1)
        else:
            topcell = layout_topcells[0]

    return topcell


def generate_klayout_switches(arguments, layout_path):
    """
    parse_switches Function that parse all the args from input to prepare switches for DRC run.

    Parameters
    ----------
    arguments : dict
        Dictionary that holds the arguments used by user in the run command. This is generated by docopt library.
    layout_path : string
        Path to the layout file that we will run DRC on.

    Returns
    -------
    dict
        Dictionary that represent all run switches passed to klayout.
    """
    switches = {}

    # No. of threads
    thrCount = 2 if arguments["--thr"] is None else int(arguments["--thr"])
    switches["thr"] = str(int(thrCount))

    if arguments["--run_mode"] in ["flat", "deep", "tiling"]:
        switches["run_mode"] = arguments["--run_mode"]
    else:
        logging.error("Allowed klayout modes are (flat , deep , tiling) only")
        exit()

    if arguments["--verbose"]:
        switches["verbose"] = "true"
    else:
        switches["verbose"] = "false"

    switches["topcell"] = get_run_top_cell_name(arguments, layout_path)
    switches["input"] = layout_path

    return switches


def check_klayout_version() -> None:
    """
    check_klayout_version checks klayout version and makes sure it would work with the DRC.
    """
    # ======= Checking Klayout version =======
    klayout_v_ = os.popen("klayout -b -v").read()
    klayout_v_ = klayout_v_.split("\n")[0]
    klayout_v_list = []

    if klayout_v_ == "":
        logging.error(
            f"Klayout is not found. Please make sure klayout is installed. Current version: {klayout_v_}"
        )
        exit(1)
    else:
        klayout_v_list = [int(v) for v in klayout_v_.split(" ")[-1].split(".")]

    if len(klayout_v_list) < 1 or len(klayout_v_list) > 3:
        logging.error(
            f"Was not able to get klayout version properly. Current version: {klayout_v_}"
        )
        exit(1)
    elif len(klayout_v_list) >= 2 and len(klayout_v_list) <= 3:
        if klayout_v_list[1] < 28:
            logging.error("Prerequisites at a minimum: KLayout 0.28.0")
            logging.error(
                "Using this klayout version has not been assesed in this development. Limits are unknown"
            )
            exit(1)

    logging.info(f"Your Klayout version is: {klayout_v_}")


def check_layout_path(layout_path):
    """
    check_layout_type checks if the layout provided is GDS or OAS. Otherwise, kill the process. We only support GDS or OAS now.

    Parameters
    ----------
    layout_path : string
        string that represent the path of the layout.

    Returns
    -------
    string
        string that represent full absolute layout path.
    """

    if not os.path.isfile(layout_path):
        logging.error("## GDS file path provided doesn't exist or not a file.")
        exit(1)

    if ".gds" not in layout_path and ".oas" not in layout_path:
        logging.error(
            "## Layout is not in GDSII or OASIS format. Please use gds format."
        )
        exit(1)

    return os.path.abspath(layout_path)


def build_switches_string(sws: dict):
    """
    build_switches_string Build switches string from dictionary.

    Parameters
    ----------
    sws : dict
        Dictionary that holds the Antenna switches.
    """
    switches_str = ""
    for k in sws:
        switches_str += f"-rd {k}={sws[k]} "

    return switches_str


def run_check(drc_file: str, drc_name: str, path: str, run_dir: str, sws: dict):
    """
    run_antenna_check run DRC check based on DRC file provided.

    Parameters
    ----------
    drc_file : str
        String that has the file full path to run.
    path : str
        String that holds the full path of the layout.
    run_dir : str
        String that holds the full path of the run location.
    sws : dict
        Dictionary that holds all switches that needs to be passed to the antenna checks.

    Returns
    -------
    string
        string that represent the path to the results output database for this run.

    """

    ## Using print because of the multiprocessing
    logging.info(
        "Running GENERIC TECH DRC-malformed {} checks on design {} on cell {}:".format(
            path, drc_name, sws["topcell"]
        )
    )

    layout_base_name = os.path.basename(path).split(".")[0]
    new_sws = sws.copy()
    report_path = os.path.join(run_dir, f"{layout_base_name}_{drc_name}.lyrdb")

    new_sws["report"] = report_path
    sws_str = build_switches_string(new_sws)

    run_str = f"klayout -b -r {drc_file} {sws_str}"

    check_call(run_str, shell=True)

    return report_path


def run_single_processor(
    arguments: dict,
    rule_deck_full_path: str,
    layout_path: str,
    switches: dict,
    drc_run_dir: str,
) -> None:
    """
    run_single_processor run the drc checks as single run.

    Parameters
    ----------
    arguments : dict
        Dictionary that holds the arguments passed to the run_drc script.
    rule_deck_full_path : str
        String that holds the path of the rule deck files.
    layout_path : str
        Path to the target layout.
    switches : dict
        Dictionary that holds all the switches that will be passed to klayout run.
    drc_run_dir : str
        Path to the run location.
    """

    list_res_db_files = []

    ## Generate run rule deck from template.
    drc_file = os.path.join(rule_deck_full_path, "generic_tech_malformed.drc")
    table_name = "main"

    ## Run Main DRC
    list_res_db_files.append(
        run_check(drc_file, table_name, layout_path, drc_run_dir, switches)
    )

    ## Check run
    check_drc_results(list_res_db_files)


def main(drc_run_dir: str, now_str: str, arguments: dict) -> None:
    """
    main function to run the DRC.

    Parameters
    ----------
    drc_run_dir : str
        String with absolute path of the full run dir.
    now_str : str
        String with the run name for logs.
    arguments : dict
        Dictionary that holds the arguments used by user in the run command. This is generated by docopt library.
    """

    # Check gds file existence
    if os.path.exists(arguments["--path"]):
        pass
    else:
        logging.error("The input GDS file path doesn't exist, please recheck.")
        exit()

    rule_deck_full_path = os.path.dirname(os.path.abspath(__file__))

    ## Check Klayout version
    check_klayout_version()

    ## Check if there was a layout provided.
    if not arguments["--path"]:
        logging.error("No provided gds file, please add one")
        exit(1)

    ## Check layout type
    layout_path = arguments["--path"]
    layout_path = check_layout_path(layout_path)

    ## Get run switches
    switches = generate_klayout_switches(arguments, layout_path)

    run_single_processor(
        arguments, rule_deck_full_path, layout_path, switches, drc_run_dir
    )


# ================================================================
# -------------------------- MAIN --------------------------------
# ================================================================

if __name__ == "__main__":
    # arguments
    arguments = docopt(__doc__, version="RUN DRC-malformed: 1.0")

    # logs format
    now_str = datetime.utcnow().strftime("drc_run_%Y_%m_%d_%H_%M_%S")

    if (
        arguments["--run_dir"] == "pwd"
        or arguments["--run_dir"] == ""
        or arguments["--run_dir"] is None
    ):
        drc_run_dir = os.path.join(os.path.abspath(os.getcwd()), now_str)
    else:
        drc_run_dir = os.path.abspath(arguments["--run_dir"])

    os.makedirs(drc_run_dir, exist_ok=True)

    logging.basicConfig(
        level=logging.DEBUG,
        handlers=[
            logging.FileHandler(os.path.join(drc_run_dir, f"{now_str}.log")),
            logging.StreamHandler(),
        ],
        format="%(asctime)s | %(levelname)-7s | %(message)s",
        datefmt="%d-%b-%Y %H:%M:%S",
    )

    # Calling main function
    main(drc_run_dir, now_str, arguments)



from __future__ import annotations

from gdsfactory.components.bend_s import bend_s
from gdsfactory.port import Port
from gdsfactory.routing.sort_ports import sort_ports as sort_ports_function
from gdsfactory.typings import Route


def get_bundle_sbend(
    ports1: Port,
    ports2: Port,
    sort_ports: bool = True,
    enforce_port_ordering: bool = True,
    **kwargs,
) -> list[Route]:
    """Returns a list of routes from ports1 to ports2.

    Args:
        ports1: start ports.
        ports2: end ports.
        sort_ports: sort ports.
        enforce_port_ordering: enforces port ordering.
        kwargs: cross_section settings.

    Returns:
        list of routes.

    """
    if sort_ports:
        ports1, ports2 = sort_ports_function(
            ports1, ports2, enforce_port_ordering=enforce_port_ordering
        )

    routes = []

    for p1, p2 in zip(ports1, ports2):
        ysize = p2.center[1] - p1.center[1]
        xsize = p2.center[0] - p1.center[0]
        bend = bend_s(size=(xsize, ysize), **kwargs)
        sbend = bend.ref()
        sbend.connect("o1", p1)
        routes.append(
            Route(
                references=[sbend],
                ports=tuple(sbend.get_ports_list()),
                length=bend.info["length"],
            )
        )
    return routes


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("test_get_route_sbend")
    pitch = 2.0
    ys_left = [0, 10, 20]
    N = len(ys_left)
    y0 = -10
    ys_right = [(i - N / 2) * pitch + y0 for i in range(N)]

    layer = (1, 0)
    right_ports = [
        gf.Port(
            f"R_{i}", center=(0, ys_right[i]), width=0.5, orientation=180, layer=layer
        )
        for i in range(N)
    ]
    left_ports = [
        gf.Port(
            f"L_{i}", center=(-50, ys_left[i]), width=0.5, orientation=0, layer=layer
        )
        for i in range(N)
    ]
    left_ports.reverse()

    routes = gf.routing.get_bundle(right_ports, left_ports, with_sbend=False)
    for route in routes:
        c.add(route.references)
    c.show(show_ports=True)



"""
In a multiple-topology Clements Scheme we can implement any universal photonic function.
"""
from __future__ import annotations

import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.mzi import mzi2x2_2x2
from gdsfactory.components.straight import straight
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing import get_route


def find_largest_component(component_list: list) -> Component:
    # TODO optimise speed here
    largest_component = component_list[0]
    for component_i in component_list:
        # If element is larger than largest component
        if (component_i.xsize * component_i.ysize) > (
            largest_component.xsize * largest_component.ysize
        ):
            largest_component = component_i
    return largest_component


@cell
def component_lattice_generic(
    network: list[list] | None = None,
) -> Component:
    """
    The shape of the `network` matrix determines the physical interconnection.
    Note that there should be at least S+1=N modes
    based on this formalism of interconnection,
    and the position of the component implements a connectivity in between the modes,
    and assumes a 2x2 network encoding.
    One nice functionality by this component is that it can generate a
    component lattice for generic variable components with different x and y pitches.
    Initially this will maximise the surface area required
    but different placement algorithms can compact the size.

    Args:
        network: A list of lists of components that are to be placed in the lattice.

    Returns:
        Component: A component lattice that implements the physical network.

    The placement matrix is in this form:
    .. math::

        M = X & 0 & X
            0 & P & 0
            X & 0 & X


    :include-source:
        import gdsfactory as gf
        from gdsfactory.components.mzi import mzi2x2_2x2

        example_component_lattice = [
            [mzi2x2_2x2(), 0, mzi2x2_2x2()],
            [0, mzi2x2_2x2(delta_length=30.0), 0],
            [mzi2x2_2x2(), 0, mzi2x2_2x2()],
        ]
        c = gf.components.component_lattice_generic(example_component_lattice)


    Another example that demonstrates the generic-nature of this component lattice
    algorithm can be with an mixed set of actively driven and passiver interferometers.
    The placement matrix is in this form:

    .. math::

        M = Y & 0 & A
            0 & B & 0
            C & 0 & Y

    :include-source:
        import gdsfactory as gf
        from gdsfactory.components import mzi2x2_2x2_phase_shifter, mzi2x2_2x2

        example_mixed_component_lattice = [
            [mzi2x2_2x2_phase_shifter(), 0, mzi2x2_2x2(delta_length=20.0)],
            [0, mzi2x2_2x2(delta_length=30.0), 0],
            [mzi2x2_2x2(delta_length=15.0), 0, mzi2x2_2x2_phase_shifter()],
        ]
        c = gf.components.component_lattice_generic(
            network=example_mixed_component_lattice
        )

    # TODO implement balanced waveguide paths function per stage
    # TODO automatic electrical fanout?
    # TODO multiple placement optimization algorithms.
    """

    network = network or [
        [mzi2x2_2x2(), 0, mzi2x2_2x2()],
        [0, mzi2x2_2x2(delta_length=30.0), 0],
        [mzi2x2_2x2(), 0, mzi2x2_2x2()],
    ]

    element_references = list()
    network = np.array(network)
    # Check number of dimensions is 2
    if network.ndim != 2:
        # Get the length and then width of the array
        raise AttributeError(
            "Physical network dimensions don't work."
            "Check the dimensional structure of your network matrix."
        )

    C = Component()
    # Estimate the size of the network fabric
    elements_list = np.vstack([np.nonzero(network), network[np.nonzero(network)]])
    largest_component = find_largest_component(elements_list[2])  # List of elements
    component_column_amount = len(network[0])
    mode_amount = component_column_amount + 1
    inter_stage_clearance_x_offset = 40
    inter_stage_clearance_y_offset = 40
    x_length = (
        mode_amount * largest_component.xsize
        + mode_amount * inter_stage_clearance_x_offset
    )
    y_length = (
        mode_amount * largest_component.ysize
        + mode_amount * inter_stage_clearance_y_offset
    )
    x_component_pitch = x_length / mode_amount
    y_component_pitch = y_length / mode_amount
    x_mode_pitch = x_length / mode_amount
    y_mode_pitch = y_length / mode_amount
    # each distinct operation on the network is a separate iteration for-loop so new
    # functionality can be extended and easily identified.

    # Create all the waveguides inputs and outputs
    # Originally implemented in gdsfactory array but got netlist errors
    interconnection_ports_array = []
    for column_j in range(mode_amount):
        interconnection_ports_array.append([])
        for row_i in range(mode_amount):
            straight_i = C << straight(length=1, width=0.5)
            interconnection_ports_array[column_j].extend([straight_i])
            interconnection_ports_array[column_j][row_i].move(
                destination=(x_mode_pitch * column_j, -y_mode_pitch * row_i)
            )

            if column_j == 0:
                # Inputs
                C.add_port(
                    port=straight_i.ports["o1"],
                    name="in_o_" + str(row_i),
                )
            elif column_j == (mode_amount - 1):
                # Outputs
                C.add_port(
                    port=straight_i.ports["o2"],
                    name="out_o_" + str(row_i),
                )

    # Place the components in between the corresponding waveguide modes
    j = 0
    k = 0
    for column_j in network:
        i = 0
        for element_i in column_j:
            # Check if element is nonzero
            if element_i != 0:
                element_references.append(C << element_i)
                element_references[k].center = (0, 0)
                element_references[k].move(
                    destination=(
                        x_component_pitch * j
                        + largest_component.xsize / 2
                        + inter_stage_clearance_x_offset / 2,
                        -y_component_pitch * i - inter_stage_clearance_y_offset,
                    )
                )
                k += 1
            i += 1
        j += 1

    # Go position by position to place and connect everything, column by column
    j = 0
    k = 0
    for column_j in network:
        i = 0
        # Connect the modes together
        for element_i in column_j:
            # Row in column
            if element_i != 0:
                # Connect the adjacent input waveguide ports to the first element columns
                # if j == 0:
                route_0 = get_route(
                    interconnection_ports_array[j][i].ports["o2"],
                    element_references[k].ports["o2"],
                    radius=5,
                )
                route_i = get_route(
                    interconnection_ports_array[j][i + 1].ports["o2"],
                    element_references[k].ports["o1"],
                    radius=5,
                )
                # Connect output of the component to the component
                route_0_out = get_route(
                    interconnection_ports_array[j + 1][i].ports["o1"],
                    element_references[k].ports["o3"],
                    radius=5,
                )
                route_i_out = get_route(
                    interconnection_ports_array[j + 1][i + 1].ports["o1"],
                    element_references[k].ports["o4"],
                    radius=5,
                )
                C.add(route_0.references)
                C.add(route_i.references)
                C.add(route_0_out.references)
                C.add(route_i_out.references)
                k += 1

            elif element_i == 0:
                # When no element at junction, connect straight ahead between
                if i == 0:
                    # If at start top row then just connect top
                    route_i = get_route(
                        interconnection_ports_array[j][i].ports["o2"],
                        interconnection_ports_array[j + 1][i].ports["o1"],
                        radius=5,
                    )
                    C.add(route_i.references)
                elif i == (len(column_j) - 1):
                    # If at end then connect bottom
                    route_i = get_route(
                        interconnection_ports_array[j][i + 1].ports["o2"],
                        interconnection_ports_array[j + 1][i + 1].ports["o1"],
                        radius=5,
                    )
                    C.add(route_i.references)

                    if column_j[i - 1] != 0:
                        # If previous element nonzero then pass
                        pass
                    elif column_j[i - 1] == 0:
                        # If previous element is zero then connect top straight
                        route_i = get_route(
                            interconnection_ports_array[j][i].ports["o2"],
                            interconnection_ports_array[j + 1][i].ports["o1"],
                            radius=5,
                        )
                        C.add(route_i.references)

                elif column_j[i - 1] == 0:
                    # If previous element is zero then connect top straight
                    route_i = get_route(
                        interconnection_ports_array[j][i].ports["o2"],
                        interconnection_ports_array[j + 1][i].ports["o1"],
                        radius=5,
                    )
                    C.add(route_i.references)

                elif column_j[i - 1] != 0:
                    # If previous element nonzero then pass
                    pass
            i += 1
        j += 1

    # Append electrical ports to component to total connectivity can be constructed.
    j = 0
    k = 0
    for column_j in network:
        i = 0
        for element_i in column_j:
            # Check if element is nonzero
            if element_i != 0:
                electrical_ports_list_i = select_ports_electrical(
                    element_references[k].ports
                ).items()
                if len(electrical_ports_list_i) > 0:
                    # Electrical ports exist in component
                    for electrical_port_i in electrical_ports_list_i:
                        C.add_port(
                            port=electrical_port_i[1],
                            name=electrical_port_i[0] + "_" + str(i) + "_" + str(j),
                        )
                        # Row column notation
                k += 1
            i += 1
        j += 1
    return C


if __name__ == "__main__":
    # from gdsfactory.components.mzi import mzi2x2_2x2
    # from gdsfactory.components.mzi_phase_shifter import mzi2x2_2x2_phase_shifter

    # example_component_lattice = [
    #     [mzi2x2_2x2(), 0, mzi2x2_2x2()],
    #     [0, mzi2x2_2x2(), 0],
    #     [mzi2x2_2x2(), 0, mzi2x2_2x2()],
    # ]
    # c = component_lattice_generic(example_component_lattice)
    # c.show(show_ports=True)

    # example_mixed_component_lattice = [
    #     [mzi2x2_2x2_phase_shifter(), 0, mzi2x2_2x2(delta_length=20.0)],
    #     [0, mzi2x2_2x2(delta_length=50.0), 0],
    #     [mzi2x2_2x2(delta_length=100.0), 0, mzi2x2_2x2_phase_shifter()],
    # ]
    # c_mixed = component_lattice_generic(example_mixed_component_lattice)
    # c_mixed.show(show_ports=True)
    c = component_lattice_generic()
    c.show()



"""snaps values and coordinates to the GDS grid in nm."""
from __future__ import annotations

import warnings
from functools import partial

import numpy as np

Value = float | tuple | np.ndarray


def is_on_grid(
    x: Value,
    nm: int | None = None,
    grid_factor: int = 1,
) -> bool:
    return np.array_equal(
        snap_to_grid(x, grid_factor=grid_factor, nm=nm), np.round(x, 6)
    )


def warn_if_not_on_grid(x: Value) -> None:
    if not is_on_grid(x):
        warnings.warn(f"{x} is not on grid", stacklevel=2)


def assert_on_grid(
    x: Value,
    nm: int | None = None,
    grid_factor: int = 1,
) -> None:
    x_grid = snap_to_grid(x, nm=nm, grid_factor=grid_factor)
    if not np.isclose(x_grid, x):
        raise ValueError(f"{x} needs to be on 1nm grid and should be {x_grid}")


assert_on_1nm_grid = partial(assert_on_grid, nm=1)
assert_on_2nm_grid = partial(assert_on_grid, nm=2)


def assert_on_2x_grid(x: float) -> None:
    x_grid = snap_to_grid(x, grid_factor=2)
    if not np.isclose(x_grid, x):
        raise ValueError(f"{x} needs to be on 2x grid and should be {x_grid}")


def snap_to_grid(
    x: Value,
    nm: int | None = None,
    grid_factor: int = 1,
) -> Value:
    """snap x to grid_sizes

    Args:
        x: value to snap.
        nm: Optional grid size in nm. If None, it will use the default grid size from PDK multiplied by grid_factor.
        grid_factor: snap to grid_factor * grid_size.
    """
    from gdsfactory.pdk import get_grid_size

    nm = nm or int(get_grid_size() * 1000 * grid_factor)
    y = nm * np.round(np.asarray(x, dtype=float) * 1e3 / nm) / 1e3

    if isinstance(x, tuple):
        return tuple(y)
    elif isinstance(x, int | float | str | np.float_):
        return float(y)
    return y


snap_to_grid2x = partial(snap_to_grid, grid_factor=2)


if __name__ == "__main__":
    print(assert_on_grid(1.1e-3))
    # print(snap_to_grid(1.1e-3))
    # print(snap_to_2nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(3.1e-3))

    # print(on_1nm_grid(1.1e-3))
    # print(on_1nm_grid(1e-3))

    # print(on_2nm_grid(1.1e-3))
    # print(on_2nm_grid(1e-3))
    # print(on_2nm_grid(2e-3))



"""Serialize component settings into YAML or strings."""
from __future__ import annotations

import functools
import hashlib
import inspect
import pathlib
from collections.abc import KeysView as dict_keys
from typing import Any

import gdstk
import numpy as np
import orjson
import pydantic
import toolz
from omegaconf import DictConfig, OmegaConf

DEFAULT_SERIALIZATION_MAX_DIGITS = 3
"""By default, the maximum number of digits retained when serializing float-like arrays"""


def get_string(value: Any) -> str:
    try:
        s = orjson.dumps(
            value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_name
        ).decode()
    except TypeError as e:
        print(f"Error serializing {value!r}")
        raise e
    return s


def clean_dict(dictionary: dict) -> dict:
    return {k: clean_value_json(v) for k, v in dictionary.items()}


def complex_encoder(obj, digits=DEFAULT_SERIALIZATION_MAX_DIGITS):
    real_part = np.round(obj.real, digits)
    imag_part = np.round(obj.imag, digits)
    return {"real": real_part, "imag": imag_part}


def clean_value_json(value: Any) -> str | int | float | dict | list | bool | None:
    """Return JSON serializable object."""
    from gdsfactory.path import Path
    from gdsfactory.pdk import get_active_pdk

    active_pdk = get_active_pdk()
    include_module = active_pdk.cell_decorator_settings.include_module

    if isinstance(value, pydantic.BaseModel):
        return clean_dict(value.model_dump())

    elif hasattr(value, "get_component_spec"):
        return value.get_component_spec()

    elif isinstance(value, bool):
        return value

    elif isinstance(value, np.integer | int):
        return int(value)

    elif isinstance(value, float | np.inexact | np.float64):
        return float(np.round(value, DEFAULT_SERIALIZATION_MAX_DIGITS))

    elif isinstance(value, complex | np.complex64 | np.complex128):
        return complex_encoder(value)

    elif isinstance(value, np.ndarray):
        value = np.round(value, DEFAULT_SERIALIZATION_MAX_DIGITS)
        return orjson.loads(orjson.dumps(value, option=orjson.OPT_SERIALIZE_NUMPY))

    elif callable(value) and isinstance(value, functools.partial):
        return clean_value_partial(value, include_module)
    elif hasattr(value, "to_dict"):
        return clean_dict(value.to_dict())

    elif callable(value) and isinstance(value, toolz.functoolz.Compose):
        return [clean_value_json(value.first)] + [
            clean_value_json(func) for func in value.funcs
        ]

    elif callable(value) and hasattr(value, "__name__"):
        return (
            {"function": value.__name__, "module": value.__module__}
            if include_module
            else {"function": value.__name__}
        )

    elif isinstance(value, Path):
        return value.hash_geometry()

    elif isinstance(value, pathlib.Path):
        return value.stem

    elif isinstance(value, dict):
        return clean_dict(value.copy())

    elif isinstance(value, DictConfig):
        return clean_dict(OmegaConf.to_container(value))

    elif isinstance(value, list | tuple | set | dict_keys):
        return [clean_value_json(i) for i in value]

    elif isinstance(value, gdstk.Polygon):
        return np.round(value.points, DEFAULT_SERIALIZATION_MAX_DIGITS)

    else:
        try:
            value_json = orjson.dumps(
                value, option=orjson.OPT_SERIALIZE_NUMPY, default=clean_value_json
            )
            return orjson.loads(value_json)
        except TypeError as e:
            print(f"Error serializing {value!r}")
            raise e


def clean_value_partial(value, include_module):
    sig = inspect.signature(value.func)
    args_as_kwargs = dict(zip(sig.parameters.keys(), value.args))
    args_as_kwargs |= value.keywords
    args_as_kwargs = clean_dict(args_as_kwargs)

    func = value.func
    while hasattr(func, "func"):
        func = func.func
    v = {
        "function": func.__name__,
        "settings": args_as_kwargs,
    }
    if include_module:
        v.update(module=func.__module__)
    return v


def clean_value_name(value: Any) -> str:
    """Returns a string representation of an object."""
    # value1 = clean_value_json(value)
    return str(clean_value_json(value))


def get_hash(value: Any) -> str:
    return hashlib.md5((clean_value_name(value)).encode()).hexdigest()[:8]


if __name__ == "__main__":
    import gdsfactory as gf

    # f = partial(gf.c.straight, length=3)
    # d = clean_value_json(f)
    # print(f"{d!r}")
    # f = partial(gf.c.straight, length=3)
    # c = f()
    # d = clean_value_json(c)
    # print(d, d)

    # xs = partial(
    #     gf.cross_section.strip,
    #     width=3,
    #     add_pins=gf.partial(gf.add_pins.add_pins_inside1nm, pin_length=0.1),
    # )
    # f = partial(gf.routing.add_fiber_array, cross_section=xs)
    # c = f()
    c = gf.cross_section.strip(width=3)
    d = clean_value_json(c)
    print(get_hash(d))



from pydantic.dataclasses import dataclass

from gdsfactory.technology.layer_views import Layer


@dataclass(kw_only=True)
class ProcessStep:
    """Generic process step."""

    name: str | None


@dataclass(kw_only=True)
class Lithography(ProcessStep):
    """Simulates lithography by generating a logical on-wafer mask from one or many layers to be used in processing operations.

    (1) First, a mask is created from the layer arguments:

    if layer=None, behaviour should be:

                                                 mask opening
                                              <---------------->
                                     ----->
        |________________|                    |________________|

        0       1        2         3          0       1        2         3


    if argument_layers is provided to layers_or, for those layers:

               layer                                 mask opening
        <---------------->                   <--------------------------->
                                     ----->
        |________________|                    |_______|________|_________|
                |__________________|
        0       1        2         3          0       1        2         3
                <------------------>
                    layers_or


    if argument_layers is provided to layers_and, for those layers:

               layer                                 mask opening
        <---------------->                            <--------->
                                     ----->
        |________________|                    |       |_________|        |
                |__________________|
        0       1        2         3          0       1        2         3
                <------------------>
                      layers_and

    if argument_layers is provided to layers_diff, for those layers:

               layer                                 mask opening
        <---------------->                    <------->
                                     ----->
        |________________|                    |________|        |        |
                |__________________|
        0       1        2         3          0       1        2         3
                <------------------>
                      layers_and

    if argument_layers is provided to layers_xor, for those layers:

              layer                                  mask opening
        <---------------->                   <-------->        <--------->
                                     ----->
        |________________|                    |_______|        |_________|
                |__________________|
        0       1        2         3          0       1        2         3
                <------------------>
                     layers_xor


    (2) Convert the logical mask into a wafer mask, opening up parts of the wafer for processing:

    if positive_tone:

           mask opening             wafer mask opened
            <------>                   <------>
                                   _____      _____
                                   |   |      |   |
                        ----->     |   |      |   | mask_thickness
        ________________           |___|______|___|


    else (negative tone):

          mask opening            wafer mask NOT opened
            <------>                   <------>
                                       ________
                                       |      |
                        ----->         |      | mask_thickness
        ________________           ____|______|____


    (3) (Optional) Planarize the resist


    Args:
        layer: main layer to use as a mask for this lithography step
        layers_union (List[Layers]): other layers to use to form the mask (see diagram)
        layers_diff (List[Layers]): other layers to use to form a mask (see diagram)
        layers_intersect (List[Layers]): other layers to use to form a mask (see diagram)
        positive_tone (bool): whether to invert the resulting mask (False) or not (True)
        resist_thickness (float): resist mask thickness, used in some simulators
        planarization_height (float): height at which to "clip" the resist above the wafer
    """

    layer: Layer | None = None
    layers_or: list | None = None
    layers_diff: list | None = None
    layers_and: list | None = None
    layers_xor: list | None = None
    resist_thickness: float | None = 0
    positive_tone: bool = True
    planarization_height: float = None


@dataclass(kw_only=True)
class Grow(Lithography):
    """Simulates masking + addition of material + liftoff.

    wafer mask opened           wafer mask opened
        <------>                   <------>
                                   ________
                                   |      |
                      ----->       | mat  | thickness
    ________________           ____|______|____

    Args:
        material (str): material tag of material to add
        thickness (float): thickness to add [nm]
        type (str): of growth/deposition (isotropic, anisotropic, etc.)
        rate (float): of growth [nm/s]
    """

    thickness: float
    material: str
    type: str
    rate: float = None


@dataclass(kw_only=True)
class Etch(Lithography):
    """Simulates masking + removal of material + strip.

    wafer mask opened          wafer mask opened
        <------>                   <----->
    ________________           _____      _____
    |              |               |      |
    |     mat      |    -----> mat | etch | mat  depth
    |______________|           ____|______|____


    Args:
        material (str): material tag to etch into
        thickness (float): thickness to remove [nm]
        type (str): of etch (isotropic, anisotropic, etc.)
        rate (float): of removal [nm/s]

    """

    material: str
    depth: float
    type: str = "anisotropic"
    rate: float = None


@dataclass(kw_only=True)
class ImplantPhysical(Lithography):
    """Simulates masking + physical ion implantation + strip.

    wafer mask opened          wafer mask opened
        <------>                   <----->
    ________________          __________________
    |              |          |                |
    |              |  ----->  |    ------- <---- range (depends on energy)
    |______________|          |________________|

    Args:
        ion (str): ion tag
        energy (float): of the ions
        dose (float): in /cm^2
        tilt (float): ion angle from out-of-plane axis. in degrees. If None, uses simulator default
        twist (float): ion angle from wafer "x-axis", in degrees. If None, uses simulator default
        rotation (float): if twist is None, toggle to split the dose 4-ways between 4 cardinal twist angles (simulates substrate rotation during implantation)
    """

    ion: str
    energy: float
    dose: float
    tilt: float = None
    twist: float = None
    rotation: float = None


@dataclass(kw_only=True)
class ImplantAnalytical(Lithography):
    """Simulates masking + physical ion implantation + strip.

    wafer mask opened          wafer mask opened
        <------>                   <----->
    ________________          __________________
    |              |          |                |
    |              |  ----->  |    ------- <---- range (depends on energy)
    |______________|          |________________|

    Args:
        ion (str): ion tag
        peak_conc (float): peak concentration
        range (float): of the ions (center of distribution)
        straggle (float): of the ions (spread of distribution)
    """

    ion: str
    peak_conc: float
    range: float
    straggle: float = None
    tilt: float = None


@dataclass(kw_only=True)
class Anneal(ProcessStep):
    """Simulates thermal diffusion of impurities and healing of defects.

    Args:
        time (float)
        temperature (float): temperature

    TODO (long term): heating/cooling time profiles
    """

    time: float
    temperature: float


@dataclass(kw_only=True)
class Planarize(ProcessStep):
    """Simulates chip planarization, "clipping" the structure above some height. Does not use masking.

         __
       _|  |___
    __|        |____
    |              |  ___                     _________________
    |              |   |  height    ----->    |               |
    |______________|  _|_  z=0                |_______________|


    Args:
        depth (float): how much to remove
    """

    height: float = 0


@dataclass(kw_only=True)
class ArbitraryStep(ProcessStep):
    """Arbitrary process step, used to handle all other cases."""

    info: str



from __future__ import annotations

from collections.abc import Iterable
from inspect import getmembers, signature

from gdsfactory.typings import Component, ComponentFactory


def get_cells(modules, verbose: bool = False) -> dict[str, ComponentFactory]:
    """Returns PCells (component functions) from a module or list of modules.

    Args:
        modules: module or iterable of modules.
        verbose: prints in case any errors occur.

    """
    modules = modules if isinstance(modules, Iterable) else [modules]

    cells = {}
    for module in modules:
        for t in getmembers(module):
            if callable(t[1]) and t[0] != "partial":
                try:
                    r = signature(t[1]).return_annotation
                    if r == Component or (
                        isinstance(r, str) and r.endswith("Component")
                    ):
                        cells[t[0]] = t[1]
                except ValueError:
                    if verbose:
                        print(f"error in {t[0]}")
    return cells


if __name__ == "__main__":
    import ubcpdk

    f = get_cells(ubcpdk.components)
    print(f.keys())



from __future__ import annotations

import gdstk

import gdsfactory as gf
from gdsfactory.typings import Component, ComponentReference, LayerSpec


def boolean_polygons(
    operand1: ComponentReference | Component | gdstk.Polygon,
    operand2: ComponentReference | Component | gdstk.Polygon,
    operation: str,
    output_layer: LayerSpec = (0, 0),
    precision: float = 1e-3,
) -> list[gdstk.Polygon]:
    """Perform a boolean operation and return the list of resulting Polygons.
    See [gdstk docs](https://heitzmann.github.io/gdstk/geometry/gdstk.boolean.html#gdstk.boolean) for details.

    Args:
        operand1: polygon set A.
        operand2: polygon set B.
        operation: the name of the operation to perform, i.e. "or", "and", "not", or "xor".
        output_layer: the layer to assign the resulting polygons.
        precision: the precision used for the operation, in microns.

    Returns: a list of gdstk Polygons on the specified output layer.
    """
    layer, datatype = gf.get_layer(output_layer)

    if operand2 is None:
        operand2 = []
    if hasattr(operand1, "get_polygons"):
        operand1 = operand1.get_polygons(as_array=False)
    if hasattr(operand2, "get_polygons"):
        operand2 = operand2.get_polygons(as_array=False)

    return gdstk.boolean(
        operand1=operand1,
        operand2=operand2,
        operation=operation,
        precision=precision,
        layer=layer,
        datatype=datatype,
    )


if __name__ == "__main__":
    import time

    n = 50
    c1 = gf.c.array(gf.c.circle(radius=10), columns=n, rows=n)
    c2 = gf.c.array(gf.c.circle(radius=9), columns=n, rows=n).ref()
    c2.movex(5)

    t0 = time.time()
    p = boolean_polygons(c1, c2, operation="xor")

    t1 = time.time()
    print(t1 - t0)

    c = gf.Component("demo")
    c.add_polygon(p)
    c.show(show_ports=True)



from __future__ import annotations

from functools import partial

import flatdict
import pydantic

import gdsfactory as gf
from gdsfactory.name import clean_name
from gdsfactory.snap import snap_to_grid as snap
from gdsfactory.typings import Layer

ignore = (
    "cross_section",
    "decorator",
    "cross_section1",
    "cross_section2",
    "contact",
    "pad",
)
prefix_to_type_default = {
    "opt": "OPTICALPORT",
    "pad": "ELECTRICALPORT",
    "vertical_dc": "ELECTRICALPORT",
    "optical": "OPTICALPORT",
    "loopback": "OPTICALPORT",
}


@pydantic.validate_call
def add_label_ehva(
    component: gf.Component,
    die: str = "demo",
    prefix_to_type: dict[str, str] = prefix_to_type_default,
    layer: Layer = (66, 0),
    metadata_ignore: list[str] | None = None,
    metadata_include_parent: list[str] | None = None,
    metadata_include_child: list[str] | None = None,
) -> gf.Component:
    """Returns Component with measurement labels.

    Args:
        component: to add labels to.
        die: string.
        port_types: list of port types to label.
        layer: text label layer.
        metadata_ignore: list of settings keys to ignore.
            Works with flatdict setting:subsetting.
        metadata_include_parent: includes parent metadata.
            Works with flatdict setting:subsetting.

    """
    metadata_ignore = metadata_ignore or []
    metadata_include_parent = metadata_include_parent or []
    metadata_include_child = metadata_include_child or []

    text = f"""DIE NAME:{die}
CIRCUIT NAME:{component.name}
"""
    info = []

    metadata = component.metadata_child["changed"]
    if metadata:
        info += [
            f"CIRCUITINFO NAME: {k}, VALUE: {v}"
            for k, v in metadata.items()
            if k not in metadata_ignore and isinstance(v, int | float | str)
        ]

    metadata = flatdict.FlatDict(component.metadata["full"])
    info += [
        f"CIRCUITINFO NAME: {clean_name(k)}, VALUE: {metadata.get(k)}"
        for k in metadata_include_parent
        if metadata.get(k)
    ]

    metadata = flatdict.FlatDict(component.metadata_child["full"])
    info += [
        f"CIRCUITINFO NAME: {k}, VALUE: {metadata.get(k)}"
        for k in metadata_include_child
        if metadata.get(k)
    ]

    text += "\n".join(info)
    text += "\n"

    info = []
    if component.ports:
        for prefix, port_type_ehva in prefix_to_type.items():
            info += [
                f"{port_type_ehva} NAME: {port.name} TYPE: {port_type_ehva}, "
                f"POSITION RELATIVE:({snap(port.x)}, {snap(port.y)}),"
                f" ORIENTATION: {port.orientation}"
                for port in component.get_ports_list(prefix=prefix)
            ]
    text += "\n".join(info)

    label = gf.Label(
        text=text,
        origin=(0, 0),
        anchor="o",
        layer=layer[0],
        texttype=layer[1],
    )
    component.add(label)
    return component


if __name__ == "__main__":
    add_label_ehva_demo = partial(
        add_label_ehva,
        die="demo_die",
        metadata_include_parent=["grating_coupler:settings:polarization"],
    )

    c = gf.c.straight(length=11)
    c = gf.c.mmi2x2(length_mmi=2.2)
    c = gf.routing.add_fiber_array(
        c,
        get_input_labels_function=None,
        grating_coupler=gf.c.grating_coupler_te,
        decorator=add_label_ehva_demo,
    )

    # add_label_ehva(c, die="demo_die", metadata_include_child=["width_mmi"])
    # add_label_ehva(c, die="demo_die", metadata_include_child=[])

    print(c.labels[0])
    c.show(show_ports=True)



from __future__ import annotations

from gdsfactory.cell import cell_without_validator
from gdsfactory.component import Component, ComponentReference


@cell_without_validator
def from_dphox(device, foundry) -> Component:
    """Returns Gdsfactory Component from a dphox Device.

    Note that you need to install dphox `pip install dphox`

    https://dphox.readthedocs.io/en/latest/index.html

    Args:
        device: Dphox device.
        foundry: Dphox foundry object.

    """
    c = Component(device.name)

    for layer_name, shapely_multipolygon in device.layer_to_polys.items():
        for poly in shapely_multipolygon:
            layer = foundry.layer_to_gds_label[layer_name]
            c.add_polygon(points=poly, layer=layer)

    for ref in device.child_to_device:
        child = from_dphox(device.child_to_device[ref], foundry)
        for gds_transform in device.child_to_transform[ref][-1]:
            new_ref = ComponentReference(
                component=child,
                origin=(gds_transform.x, gds_transform.y),
                rotation=gds_transform.angle,
                magnification=gds_transform.mag,
                x_reflection=gds_transform.flip_y,
            )
            new_ref.owner = c
            c.add(new_ref)

    for port_name, port in device.port.items():
        c.add_port(
            name=port_name,
            center=(port.x, port.y),
            orientation=port.a,
            width=port.w,
            layer=foundry.layer_to_gds_label.get(port.layer, (1, 0)),
        )
    return c


if __name__ == "__main__":
    import dphox as dp
    from dphox.demo import mzi

    c = from_dphox(mzi, foundry=dp.foundry.FABLESS)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@gf.cell
def ring_single_sample(
    gap: float = 0.2,
    radius: float = 10.0,
    length_x: float = 4.0,
    length_y: float = 0.010,
    coupler_ring: ComponentSpec = "coupler_ring",
    straight: ComponentSpec = "straight",
    bend: ComponentSpec = "bend_euler",
    cross_section: CrossSectionSpec = "xs_sc",
    **kwargs,
) -> Component:
    """Single bus ring made of a ring coupler.

    (cb: bottom) connected with two vertical straights (wl: left, wr: right)
    two bends (bl, br) and horizontal straight (wg: top).

    Args:
        gap: gap between for coupler.
        radius: for the bend and coupler.
        length_x: ring coupler length.
        length_y: vertical straight length.
        coupler_ring: ring coupler function.
        straight: straight function.
        bend: 90 degrees bend function.
        cross_section: spec.
        kwargs: cross_section settings.


    .. code::

          bl-wt-br
          |      |
          wl     wr length_y
          |      |
         --==cb==-- gap

          length_x

    """
    gap = gf.snap.snap_to_grid(gap, grid_factor=2)

    xs = gf.get_cross_section(cross_section, **kwargs)

    coupler_ring_component = gf.get_component(
        coupler_ring,
        bend=bend,
        gap=gap,
        radius=radius,
        length_x=length_x,
        cross_section=xs,
    )
    straight_side = gf.get_component(straight, length=length_y, cross_section=xs)
    straight_top = gf.get_component(straight, length=length_x, cross_section=xs)

    bend = gf.get_component(bend, radius=radius, cross_section=xs)

    c = Component()
    cb = c << coupler_ring_component
    wl = c << straight_side
    wr = c << straight_side
    bl = c << bend
    br = c << bend
    wt = c << straight_top
    # wt.mirror(p1=(0, 0), p2=(1, 0))

    wl.connect(port="o2", destination=cb.ports["o2"])
    bl.connect(port="o2", destination=wl.ports["o1"])

    wt.connect(port="o2", destination=bl.ports["o1"])
    br.connect(port="o2", destination=wt.ports["o1"])
    wr.connect(port="o1", destination=br.ports["o1"])

    c.add_port("o2", port=cb.ports["o2"])
    c.add_port("o1", port=cb.ports["o1"])
    return c


def test_ring_single_sample() -> None:
    assert ring_single_sample()


if __name__ == "__main__":
    c = ring_single_sample(width=2, gap=1, layer=(2, 0))
    print(c.ports)
    c.show(show_ports=True)



from __future__ import annotations

import gdsfactory as gf
from gdsfactory import Component
from gdsfactory.cross_section import strip
from gdsfactory.typings import CrossSectionSpec


def _generate_fins(c, x, fin_size, bend):
    num_fins = int(x.width // (2 * fin_size[1]))
    y0 = bend.ports["o1"].x, -num_fins * (2 * fin_size[1]) / 2.0 + fin_size[1]

    for i in range(num_fins):
        y = y0 + i * 2 * fin_size[1]
        rectangle = c << gf.components.rectangle(
            size=(fin_size[0], fin_size[1]),
            layer=x.layer,
            centered=True,
            port_type=None,
            port_orientations=None,
        )
        rectangle.movex(
            destination=bend.ports["o1"].x
            - (1 - bend.ports["o1"].orientation / 90.0) * fin_size[0] / 2.0
        )

        rectangle.movey(
            origin=rectangle.y,
            destination=bend.ports["o1"].y + y,
        )
        c.absorb(rectangle)

    return c


def _generate_bends(c, x_top, x_bot, dx, dy, gap):
    input_bend_top = (
        c << gf.components.bend_s(size=(dx, dy), cross_section=x_top).mirror()
    )

    input_bend_bottom = c << gf.components.bend_s(
        size=(dx, dy), cross_section=x_bot
    ).mirror().mirror(p1=(1, 0))

    dy = input_bend_bottom.ports["o2"].y - input_bend_top.ports["o2"].y

    input_bend_top.movey(destination=dy / 2.0 + gap / 2.0)

    input_bend_bottom.movey(
        destination=-dy / 2.0 - (x_bot.width / 2.0 + x_top.width / 2.0 + gap / 2.0)
    )

    return (c, input_bend_top, input_bend_bottom)


def _generate_straights(c, length, x_top, x_bot, input_bend_top, input_bend_bottom):
    top_straight = c << gf.components.straight(length=length, cross_section=x_top)

    bottom_straight = c << gf.components.straight(length=length, cross_section=x_bot)

    top_straight.movey(destination=input_bend_top.ports["o2"].y)

    bottom_straight.movey(destination=input_bend_bottom.ports["o2"].y)

    return (c, top_straight, bottom_straight)


def _generate_gratings(c, length, period, dc, gap, x, bottom_straight, x_bot):
    num = length // period
    x_size = period * dc
    start = length - (num - 1) * period

    for i in range(int(num)):
        x_start = start + i * period
        rectangle = c << gf.components.rectangle(
            size=(x_size, gap),
            layer=x.layer,
            centered=True,
            port_type=None,
            port_orientations=None,
        )
        rectangle.movex(destination=x_start)
        rectangle.movey(
            destination=bottom_straight.ports["o1"].y + x_bot.width / 2.0 + gap / 2.0
        )

    return c


def _absorb(c, *refs):
    for ref in refs:
        c.absorb(ref)
    return c


@gf.cell
def cdc(
    length: float = 30.0,
    gap: float = 0.5,
    period: float = 0.220,
    dc: float = 0.5,
    dx: float = 10.0,
    dy: float = 1.8,
    width_top: float = 2.0,
    width_bot: float = 0.75,
    fins: bool = False,
    fin_size: tuple[float, float] = (0.2, 0.05),
    cross_section: CrossSectionSpec = strip,
    **kwargs,
) -> Component:
    """Grating-Assisted Contra-Directional Coupler.

    Args:
       length : Length of the coupling region.
       gap: Distance between the two straights.
       period: Period of the grating.
       dc: Duty cycle of the grating. Must be between 0 and 1.
       width_top: Width of the top straight in the coupling region.
       width_bot: Width of the bottom straight in the coupling region.
       dx: size of bends in x-direction.
       dy: size of bends in y-direction.
       fins: If `True`, adds fins to the input/output straights.
        In this case a different template for the component must be specified.
        This feature is useful when performing electron-beam lithography
        and using different beam currents
        for fine features (helps to reduce stitching errors).
       fin_size: Specifies the x- and y-size of the `fins`. Defaults to 200 nm x 50 nm.
       cross_section: CrossSection spec.

    kwargs:
        cross_section kwargs.
    """
    x = gf.get_cross_section(cross_section, **kwargs)
    x_top = x.copy(width=width_top, add_pins_function_name=None)
    x_bot = x.copy(width=width_bot, add_pins_function_name=None)

    c = gf.Component()

    c, input_bend_top, input_bend_bottom = _generate_bends(c, x_top, x_bot, dx, dy, gap)
    c, top_straight, bottom_straight = _generate_straights(
        c, length, x_top, x_bot, input_bend_top, input_bend_bottom
    )

    bend_output_top = c << input_bend_top.parent.mirror()
    bend_output_bottom = c << input_bend_bottom.parent.mirror()

    input_bend_top.connect("o2", top_straight.ports["o1"])
    input_bend_bottom.connect("o2", bottom_straight.ports["o1"])
    bend_output_top.connect("o2", top_straight.ports["o2"])
    bend_output_bottom.connect("o2", bottom_straight.ports["o2"])

    c = _generate_gratings(c, length, period, dc, gap, x, bottom_straight, x_bot)

    if fins:
        c = _generate_fins(c, x_top, fin_size, input_bend_top)
        c = _generate_fins(c, x_bot, fin_size, input_bend_bottom)
        c = _generate_fins(c, x_top, fin_size, bend_output_top)
        c = _generate_fins(c, x_bot, fin_size, bend_output_bottom)

    c = _absorb(
        c,
        input_bend_top,
        input_bend_bottom,
        top_straight,
        bottom_straight,
        bend_output_top,
        bend_output_bottom,
    )

    x.add_bbox(c)
    x.add_pins(c)

    c.add_port("o1", port=input_bend_top.ports["o1"])
    c.add_port("o2", port=input_bend_bottom.ports["o1"])
    c.add_port("o3", port=bend_output_top.ports["o1"])
    c.add_port("o4", port=bend_output_bottom.ports["o1"])
    return c


if __name__ == "__main__":
    c = cdc(fins=False)
    c.show(show_ports=False)



"""Adiabatic tapers from CSV files."""
from __future__ import annotations

import pathlib
from functools import partial
from pathlib import Path

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.typings import CrossSectionSpec

data = pathlib.Path(__file__).parent / "csv_data"


@gf.cell
def taper_from_csv(
    filepath: Path = data / "taper_strip_0p5_3_36.csv",
    cross_section: CrossSectionSpec = "xs_sc",
) -> Component:
    """Returns taper from CSV file.

    Args:
        filepath: for CSV file.
        cross_section: specification (CrossSection, string, CrossSectionFactory dict).
    """
    import pandas as pd

    taper_data = pd.read_csv(filepath)
    xs = taper_data["x"].values * 1e6
    ys = np.round(taper_data["width"].values * 1e6 / 2.0, 3)

    x = gf.get_cross_section(cross_section)
    layer = x.layer

    c = gf.Component()
    c.add_polygon(list(zip(xs, ys)) + list(zip(xs, -ys))[::-1], layer=layer)

    # for cladding_layer, cladding_offset in zip(x.cladding_layers, x.cladding_offsets):
    #     ys_trench = ys + cladding_offset
    #     c.add_polygon(
    #         list(zip(xs, ys_trench)) + list(zip(xs, -ys_trench))[::-1],
    #         layer=cladding_layer,
    #     )

    c.add_port(
        name="o1",
        center=(xs[0], 0),
        width=2 * ys[0],
        orientation=180,
        layer=layer,
        cross_section=x,
    )
    c.add_port(
        name="o2",
        center=(xs[-1], 0),
        width=2 * ys[-1],
        orientation=0,
        layer=layer,
        cross_section=x,
    )
    if x.add_bbox:
        c = x.add_bbox(c)
    if x.add_pins:
        c = x.add_pins(c)
    return c


taper_0p5_to_3_l36 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_3_36.csv")
taper_w10_l100 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_10_100.csv")
taper_w10_l150 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_10_150.csv")
taper_w10_l200 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_10_200.csv")
taper_w11_l200 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_11_200.csv")
taper_w12_l200 = partial(taper_from_csv, filepath=data / "taper_strip_0p5_12_200.csv")


if __name__ == "__main__":
    # c = taper_0p5_to_3_l36()
    c = taper_w10_l100()
    # c = taper_w11_l200()
    c.show(show_ports=True)



"""GDS regression test. Inspired by lytest."""
import filecmp
import pathlib
import shutil

import gdsfactory as gf
from gdsfactory.config import CONF, PATH, logger
from gdsfactory.name import clean_name


class GeometryDifference(Exception):
    pass


PathType = pathlib.Path | str


def diff(
    ref_file: PathType, run_file: PathType, xor: bool = True, test_name: str = ""
) -> bool:
    """Returns True if files are different, prints differences and shows them in klayout.

    Args:
        ref_file: reference (old) file.
        run_file: run (new) file.
        xor: runs xor on every layer between ref and run files.
        test_name: prefix for the new cell.
    """
    try:
        from kfactory import KCell, kdb
    except ImportError as e:
        print(
            "You can install `pip install gdsfactory[kfactory]` for using maskprep. "
            "And make sure you use python >= 3.10"
        )
        raise e
    ref = read_top_cell(ref_file)
    run = read_top_cell(run_file)
    ld = kdb.LayoutDiff()

    a_regions: dict[int, kdb.Region] = {}
    a_texts: dict[int, kdb.Texts] = {}
    b_regions: dict[int, kdb.Region] = {}
    b_texts: dict[int, kdb.Texts] = {}

    def get_region(key, regions: dict[int, kdb.Region]) -> kdb.Region:
        if key not in regions:
            reg = kdb.Region()
            regions[key] = reg
            return reg
        else:
            return regions[key]

    def get_texts(key, texts_dict: dict[int, kdb.Texts]) -> kdb.Texts:
        if key not in texts_dict:
            texts = kdb.Texts()
            texts_dict[key] = texts
            return texts
        else:
            return texts_dict[key]

    def polygon_diff_a(anotb: kdb.Polygon, prop_id: int):
        get_region(ld.layer_index_a(), a_regions).insert(anotb)

    def polygon_diff_b(bnota: kdb.Polygon, prop_id: int):
        get_region(ld.layer_index_b(), b_regions).insert(bnota)

    def cell_diff_a(cell: kdb.Cell):
        print(f"{cell.name} only in old")

    def cell_diff_b(cell: kdb.Cell):
        print(f"{cell.name} only in new")

    def text_diff_a(anotb: kdb.Text, prop_id: int):
        get_texts(ld.layer_index_a(), a_texts).insert(anotb)

    def text_diff_b(bnota: kdb.Text, prop_id: int):
        get_texts(ld.layer_index_b(), b_texts).insert(bnota)

    ld.on_cell_in_a_only = lambda anotb: cell_diff_a(anotb)
    ld.on_cell_in_b_only = lambda anotb: cell_diff_b(anotb)
    ld.on_text_in_a_only = lambda anotb, prop_id: text_diff_a(anotb, prop_id)
    ld.on_text_in_b_only = lambda anotb, prop_id: text_diff_b(anotb, prop_id)

    ld.on_polygon_in_a_only = lambda anotb, prop_id: polygon_diff_a(anotb, prop_id)
    ld.on_polygon_in_b_only = lambda anotb, prop_id: polygon_diff_b(anotb, prop_id)

    if CONF.difftest_ignore_cell_name_differences:
        ld.on_cell_name_differs = lambda anotb: print(f"cell name differs {anotb.name}")
        equal = ld.compare(
            ref._kdb_cell, run._kdb_cell, kdb.LayoutDiff.SmartCellMapping, 1
        )
    else:
        equal = ld.compare(ref._kdb_cell, run._kdb_cell, kdb.LayoutDiff.Verbose, 1)

    if not equal:
        c = KCell(f"{test_name}_difftest")
        refdiff = KCell(f"{test_name}_old")
        rundiff = KCell(f"{test_name}_new")

        refdiff.copy_tree(ref._kdb_cell)
        rundiff.copy_tree(run._kdb_cell)
        _ = c << refdiff
        _ = c << rundiff

        if xor:
            diff = KCell(f"{test_name}_xor")

            for layer in c.kcl.layer_infos():
                if layer in run.kcl.layer_infos() and layer in ref.kcl.layer_infos():
                    layer_ref = ref.layer(layer)
                    layer_run = run.layer(layer)

                    region_run = kdb.Region(run.begin_shapes_rec(layer_run))
                    region_ref = kdb.Region(ref.begin_shapes_rec(layer_ref))
                    region_diff = region_run - region_ref

                    if not region_diff.is_empty():
                        layer_id = c.layer(layer)
                        region_xor = region_ref ^ region_run
                        diff.shapes(layer_id).insert(region_xor)
                        is_sliver = region_xor.sized(-1).is_empty()
                        message = f"{test_name}: XOR difference on layer {layer}"
                        if is_sliver:
                            message += " (sliver or label)"
                        print(message)
                elif layer in run.kcl.layer_infos():
                    layer_id = run.layer(layer)
                    region = kdb.Region(run.begin_shapes_rec(layer_id))
                    diff.shapes(layer_id).insert(region)
                    print(f"{test_name}: layer {layer} only exists in updated cell")
                elif layer in ref.kcl.layer_infos():
                    layer_id = ref.layer(layer)
                    region = kdb.Region(ref.begin_shapes_rec(layer_id))
                    diff.shapes(layer_id).insert(region)
                    print(f"{test_name}: layer {layer} missing from updated cell")

            _ = c << diff

        c.show()
        return True
    return False


def difftest(
    component: gf.Component,
    test_name: gf.Component | None = None,
    dirpath: pathlib.Path = PATH.gds_ref,
    xor: bool = True,
    dirpath_run: pathlib.Path = PATH.gds_run,
) -> None:
    """Avoids GDS regressions tests on the GeometryDifference.

    If files are the same it returns None. If files are different runs XOR
    between new component and the GDS reference stored in dirpath and
    raises GeometryDifference if there are differences and show differences in KLayout.

    If it runs for the fist time it just stores the GDS reference.

    Args:
        component: to test if it has changed.
        test_name: used to store the GDS file.
        dirpath: directory where reference files are stored.
        xor: runs XOR.
        dirpath_run: directory to store gds file generated by the test.
    """
    test_name = test_name or (
        f"{component.function_name}_{component.name}"
        if hasattr(component, "function_name")
        and component.name != component.function_name
        else f"{component.name}"
    )
    filename = f"{test_name}.gds"
    dirpath_ref = dirpath
    dirpath_ref.mkdir(exist_ok=True, parents=True)
    dirpath_run.mkdir(exist_ok=True, parents=True)

    ref_file = dirpath_ref / f"{clean_name(test_name)}.gds"
    run_file = dirpath_run / filename

    component = gf.get_component(component)
    run_file = component.write_gds(gdspath=run_file)

    if not ref_file.exists():
        shutil.copy(run_file, ref_file)
        raise AssertionError(
            f"Reference GDS file for {test_name!r} not found. Writing to {ref_file!r}"
        )

    if filecmp.cmp(ref_file, run_file, shallow=False):
        return

    if diff(ref_file=ref_file, run_file=run_file, xor=xor, test_name=test_name):
        print(
            f"\ngds_run {filename!r} changed from gds_ref {str(ref_file)!r}\n"
            "You can check the differences in Klayout GUI or run XOR with\n"
            f"gf gds diff --xor {ref_file} {run_file}\n"
        )
        try:
            overwrite(ref_file, run_file)
        except OSError as exc:
            raise GeometryDifference(
                "\n"
                f"{filename!r} changed from reference {str(ref_file)!r}. "
                "Run `pytest -s` to step and check differences in klayout GUI."
            ) from exc


def overwrite(ref_file, run_file):
    val = input("Save current GDS as the new reference (Y)? [Y/n]")
    if val.upper().startswith("N"):
        raise GeometryDifference

    logger.info(f"deleting file {str(ref_file)!r}")
    ref_file.unlink()
    shutil.copy(run_file, ref_file)
    raise GeometryDifference


def read_top_cell(arg0):
    from kfactory import KCLayout

    kcl = KCLayout()
    kcl.read(arg0)
    return kcl[kcl.top_cell().name]


if __name__ == "__main__":
    # print([i.name for i in c.get_dependencies()])
    # c.show()
    # c.name = "mzi"
    c = gf.components.straight(length=20, layer=(1, 0))
    c.name = "a:demo"
    c.show()
    difftest(c, "straight", dirpath=PATH.cwd)

    # component = gf.components.mzi()
    # test_name = "mzi"
    # filename = f"{test_name}.gds"
    # dirpath = PATH.cwd
    # dirpath_ref = dirpath / "gds_ref"
    # dirpath_run = GDSDIR_TEMP

    # ref_file = dirpath_ref / f"{test_name}.gds"
    # run_file = dirpath_run / filename
    # run = gf.get_component(component)
    # run_file = run.write_gds(gdspath=run_file)

    # if not ref_file.exists():
    #     component.write_gds(gdspath=ref_file)
    #     raise AssertionError(
    #         f"Reference GDS file for {test_name!r} not found. Writing to {ref_file!r}"
    #     )

    # ref = read_top_cell(ref_file)
    # run = read_top_cell(run_file)
    # ld = kdb.LayoutDiff()

    # print(ld.compare(ref._kdb_cell, run._kdb_cell))



from __future__ import annotations

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.taper_cross_section import taper_cross_section
from gdsfactory.cross_section import strip
from gdsfactory.typings import CrossSectionSpec, LayerSpecs


@gf.cell
def terminator(
    length: float | None = 50,
    cross_section_input: CrossSectionSpec = strip,
    cross_section_tip: CrossSectionSpec | None = None,
    tapered_width: float = 0.2,
    doping_layers: LayerSpecs = ("NPP",),
    doping_offset: float = 1.0,
) -> gf.Component:
    """Returns doped taper to terminate waveguides.

    Args:
        length: distance between input and narrow tapered end.
        cross_section_input: input cross-section.
        cross_section_tip: cross-section at the end of the termination.
        tapered_width: width of the default cross-section at the end of the termination.
            Only used if cross_section_tip is not None.
        doping_layers: doping layers to superimpose on the taper. Default N++.
        doping_offset: offset of the doping layer beyond the bbox
    """
    c = Component()

    cross_section_tip = cross_section_tip or gf.get_cross_section(
        cross_section_input, width=tapered_width
    )

    taper = c << gf.get_component(
        taper_cross_section,
        length=length,
        cross_section1=cross_section_input,
        cross_section2=cross_section_tip,
    )

    for layer in doping_layers:
        _ = c << gf.components.bbox(
            bbox=taper.bbox, layer=layer, top=doping_offset, bottom=doping_offset
        )

    c.add_port(name="o1", port=taper.ports["o1"])
    return c


if __name__ == "__main__":
    # c = terminator(cross_section_input=partial(gf.cross_section.strip, width=10))
    c = terminator()
    c.show(show_ports=True)



import warnings

from gdsfactory.component import Component, ComponentReference, Port
from gdsfactory.typings import CrossSectionSpec


def taper_to_cross_section(
    port: Port, cross_section: CrossSectionSpec
) -> ComponentReference | None:
    """Returns taper ComponentReference from a port to a given cross-section \
            placed so that it connects to the input port.

    Assumes that the taper component has `width1` and `width2` which map to the input and output port widths.

    Args:
        port: a port to connect to, usually from a ComponentReference
        cross_section: a cross-section to transition to

    .. plot::
        :include-source:

        from gdsfactory.routing.auto_taper import taper_to_cross_section
        from gdsfactory.cross_section import strip
        import gdsfactory as gf

        c = gf.Component()

        # create a component reference to connect to
        wg = c << gf.components.straight()

        # create a taper reference transitioning to strip from the rib waveguide
        taper = taper_to_cross_section(wg.ports['o1'], strip(width=2.0))

        # add the taper reference to the parent component
        c.add(taper)
        c.plot()
    """
    from gdsfactory.pdk import (
        get_active_pdk,
        get_component,
        get_cross_section,
        get_layer,
    )

    port_layer = get_layer(port.layer)
    port_width = port.width
    cross_section = get_cross_section(cross_section)
    cs_layer = get_layer(cross_section.layer)
    cs_width = cross_section.width
    layer_transitions = get_active_pdk().layer_transitions

    if port_layer != cs_layer:
        try:
            taper_name = layer_transitions[(port_layer, cs_layer)]
        except KeyError as e:
            raise KeyError(
                f"No registered tapers between routing layers {port_layer} and {cs_layer}!"
            ) from e
    elif abs(port_width - cs_width) > 0.001:
        try:
            taper_name = layer_transitions[port_layer]
        except KeyError:
            warnings.warn(
                f"No registered width taper for layer {port_layer}. Skipping."
            )
            return None
    else:
        return None
    taper = get_component(taper_name, width1=port_width, width2=cs_width)
    input_port_name = _get_taper_io_port_names(component=taper)[0]
    return ComponentReference(taper).connect(input_port_name, port)


def _get_taper_io_port_names(component: Component) -> list[str]:
    # this is kind of a hack, but o1 < o2, in0 < out0... hopefully nobody has any other wacky conventions!
    return sorted(component.ports.keys())


def _auto_taper(
    routing_func,
    ports1,
    ports2,
    cross_section: CrossSectionSpec,
    auto_taper=True,
    **kwargs,
):
    from gdsfactory.pdk import get_cross_section

    if not auto_taper:
        return routing_func(ports1, ports2, cross_section=cross_section, **kwargs)
    tapers = []
    inner_ports = []
    ports_original = [ports1, ports2]
    x = get_cross_section(cross_section)

    for port_group in ports_original:
        new_port_group = []
        taper_group = []
        for port in port_group:
            taper = taper_to_cross_section(port, x)
            taper_group.append(taper)
            if taper is None:
                new_port_group.append(port)
            else:
                new_port_group.append(taper.ports["out0"])
        inner_ports.append(new_port_group)
        tapers.append(taper_group)
    routes = routing_func(
        inner_ports[0], inner_ports[1], cross_section=cross_section, **kwargs
    )
    for route, port1, port2, taper1, taper2 in zip(
        routes, ports1, ports2, tapers[0], tapers[1]
    ):
        if taper1:
            route.references.append(taper1)
        if taper2:
            route.references.append(taper2)
        route.ports = (port1, port2)
    return routes



from __future__ import annotations

import numpy as np

from gdsfactory.component import Component
from gdsfactory.typings import Floats, Layers


def to_np(
    component: Component,
    nm_per_pixel: int = 20,
    layers: Layers = ((1, 0),),
    values: Floats | None = None,
    pad_width: int = 1,
) -> np.ndarray:
    """Returns a pixelated numpy array from Component polygons.

    Args:
        component: Component.
        nm_per_pixel: you can go from 20 (coarse) to 4 (fine).
        layers: to convert. Order matters (latter overwrite former).
        values: associated to each layer (defaults to 1).
        pad_width: padding pixels around the image.

    """
    import skimage.draw as skdraw

    pixels_per_um = (1 / nm_per_pixel) * 1e3
    xmin, ymin = component.bbox[0]
    xmax, ymax = component.bbox[1]
    shape = (
        int(np.ceil(xmax - xmin) * pixels_per_um),
        int(np.ceil(ymax - ymin) * pixels_per_um),
    )
    img = np.zeros(shape, dtype=float)
    layer_to_polygons = component.get_polygons(by_spec=True, depth=None)

    values = values or [1] * len(layers)

    for layer, value in zip(layers, values):
        if layer in layer_to_polygons:
            polygons = layer_to_polygons[layer]
            for polygon in polygons:
                r = polygon[:, 0] - xmin
                c = polygon[:, 1] - ymin
                rr, cc = skdraw.polygon(
                    r * pixels_per_um, c * pixels_per_um, shape=shape
                )
                img[rr, cc] = value

    return np.pad(img, pad_width=pad_width)


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    import gdsfactory as gf

    c = gf.components.straight()
    c = gf.components.bend_circular()
    # i = to_np(c, nm_per_pixel=250)
    i = to_np(c, nm_per_pixel=20)
    c.show(show_ports=True)
    plt.imshow(i.transpose(), origin="lower")
    plt.colorbar()
    plt.show()



"""Add ports from pin markers or labels."""
from __future__ import annotations

from functools import partial

import numpy as np
from numpy import arctan2, degrees, isclose

from gdsfactory.component import Component
from gdsfactory.port import Port, read_port_markers, sort_ports_clockwise
from gdsfactory.snap import snap_to_grid
from gdsfactory.typings import LayerSpec


def add_ports_from_markers_square(
    component: Component,
    pin_layer: LayerSpec = "DEVREC",
    port_layer: LayerSpec | None = None,
    orientation: int | None = 90,
    min_pin_area_um2: float = 0,
    max_pin_area_um2: float = 150 * 150,
    pin_extra_width: float = 0.0,
    port_names: tuple[str, ...] | None = None,
    port_name_prefix: str | None = None,
    port_type: str = "optical",
) -> Component:
    """Add ports from square markers at the port center in port_layer.

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: for port markers.
        port_layer: for the new created port.
        orientation: in degrees 90 north, 0 east, 180 west, 270 south.
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2.
        max_pin_area_um2: ignore pins for area above certain size.
        pin_extra_width: 2*offset from pin to straight.
        port_names: names of the ports (defaults to {i}).
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical.
        port_type: optical, electrical.
    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    port_markers = read_port_markers(component, (pin_layer,))
    port_names = port_names or [
        f"{port_name_prefix}{i+1}" for i in range(len(port_markers.polygons))
    ]
    layer = port_layer or pin_layer

    for port_name, p in zip(port_names, port_markers.polygons):
        (xmin, ymin), (xmax, ymax) = p.bounding_box()
        x, y = np.sum(p.bounding_box(), 0) / 2

        dy = snap_to_grid(ymax - ymin)
        dx = snap_to_grid(xmax - xmin)
        if dx == dy and max_pin_area_um2 > dx * dy > min_pin_area_um2:
            x = x
            y = y
            component.add_port(
                port_name,
                center=(x, y),
                width=dx - pin_extra_width,
                orientation=orientation,
                layer=layer,
            )
    return component


def add_ports_from_markers_center(
    component: Component,
    pin_layer: LayerSpec = "PORT",
    port_layer: LayerSpec | None = None,
    inside: bool = False,
    tol: float = 0.1,
    pin_extra_width: float = 0.0,
    min_pin_area_um2: float | None = None,
    max_pin_area_um2: float = 150.0 * 150.0,
    skip_square_ports: bool = False,
    xcenter: float | None = None,
    ycenter: float | None = None,
    port_name_prefix: str | None = None,
    port_type: str = "optical",
    short_ports: bool = False,
    auto_rename_ports: bool = True,
    debug: bool = False,
) -> Component:
    """Add ports from pins guessing port orientation from component boundary.

    Args:
        component: to read polygons from and to write ports to.
        pin_layer: GDS layer for maker [int, int].
        port_layer: for the new created port.
        inside: True-> markers  inside. False-> markers at center.
        tol: tolerance area to search ports at component boundaries xmin, ymin, xmax, xmax.
        pin_extra_width: 2*offset from pin to straight.
        min_pin_area_um2: ignores pins with area smaller than min_pin_area_um2.
        max_pin_area_um2: ignore pins for area above certain size.
        skip_square_ports: skips square ports (hard to guess orientation).
        xcenter: for guessing orientation of rectangular ports.
        ycenter: for guessing orientation of rectangular ports.
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical ports.
        port_type: type of port (optical, electrical ...).
        short_ports: if the port is on the short side rather than the long side
        auto_rename_ports:
        debug: if True prints ports that are skipped.

    For inside=False the port location is at the middle of the PIN

    .. code::
           _______________
          |               |
          |               |
         |||             |||____  | pin_extra_width/2 > 0
         |||             |||
         |||             |||____
         |||             |||
          |      __       |
          |_____|__|______|
                |__|

    For inside=True all the pin is inside the port

    .. code::
           _______________
          |               |
          |               |
          |_              |
          | |             |
          |_|             |
          |               |
          |      __       |
          |_____|__|______|

    dx < dy: port is east or west
        x > xc: east
        x < xc: west

    dx > dy: port is north or south
        y > yc: north
        y < yc: south

    dx = dy
        x > xc: east
        x < xc: west
    """
    xc = xcenter or component.x
    yc = ycenter or component.y
    xmax = component.xmax
    xmin = component.xmin
    ymax = component.ymax
    ymin = component.ymin

    port_markers = read_port_markers(component, layers=(pin_layer,))
    layer = port_layer or pin_layer
    port_locations = []

    ports = {}

    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default

    for i, p in enumerate(port_markers.polygons):
        port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else str(i)
        (pxmin, pymin), (pxmax, pymax) = p.bounding_box()
        x, y = np.sum(p.bounding_box(), 0) / 2

        dy = pymax - pymin
        dx = pxmax - pxmin

        if min_pin_area_um2 and dx * dy < min_pin_area_um2:
            if debug:
                print(f"skipping port at ({x}, {y}) with min_pin_area_um2 {dx * dy}")
            continue

        if max_pin_area_um2 and dx * dy > max_pin_area_um2:
            continue

        if skip_square_ports and snap_to_grid(dx) == snap_to_grid(dy):
            if debug:
                print(f"skipping square port at ({x}, {y})")
            continue

        orientation = -1

        # rectangular ports orientation is easier to detect
        if dy < dx if short_ports else dx < dy:
            if x > xc:  # east
                orientation = 0
                width = dy
                x = pxmax if inside else x
            elif x < xc:  # west
                orientation = 180
                width = dy
                x = pxmin if inside else x
        elif dy > dx if short_ports else dx > dy:
            if y > yc:  # north
                orientation = 90
                width = dx
                y = pymax if inside else y
            elif y < yc:  # south
                orientation = 270
                width = dx
                y = pymin if inside else y

        # square ports ports are harder to detect orientation
        elif pxmax > xmax - tol:  # east
            orientation = 0
            width = dy
            x = pxmax if inside else x
        elif pxmin < xmin + tol:  # west
            orientation = 180
            width = dy
            x = pxmin if inside else x
        elif pymax > ymax - tol:  # north
            orientation = 90
            width = dx
            y = pymax if inside else y
        elif pymin < ymin + tol:  # south
            orientation = 270
            width = dx
            y = pymin if inside else y

        elif pxmax > xc:
            orientation = 0
            width = dy
            x = pxmax if inside else x

        elif pxmax < xc:
            orientation = 180
            width = dy
            x = pxmin if inside else x

        if orientation == -1:
            raise ValueError(f"Unable to detector port at ({x}, {y})")

        x = snap_to_grid(x)
        y = snap_to_grid(y)
        width = np.round(width - pin_extra_width, 3)

        if (x, y) not in port_locations:
            port_locations.append((x, y))
            ports[port_name] = Port(
                name=port_name,
                center=(x, y),
                width=width,
                orientation=orientation,
                layer=layer,
                port_type=port_type,
            )

    ports = sort_ports_clockwise(ports)

    for port_name, port in ports.items():
        if port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )

        else:
            component.add_port(name=port_name, port=port)
    if auto_rename_ports:
        component.auto_rename_ports()
    return component


add_ports_from_markers_inside = partial(add_ports_from_markers_center, inside=True)


def add_ports_from_labels(
    component: Component,
    port_width: float,
    port_layer: LayerSpec,
    xcenter: float | None = None,
    port_name_prefix: str | None = None,
    port_type: str = "optical",
    get_name_from_label: bool = False,
    layer_label: LayerSpec | None = None,
    fail_on_duplicates: bool = False,
    port_orientation: float | None = None,
    guess_port_orientation: bool = True,
) -> Component:
    """Add ports from labels.

    Assumes that all ports have a label at the port center.
    because labels do not have width, you have to manually specify the ports width

    Args:
        component: to read polygons from and to write ports to.
        port_width: for ports.
        port_layer: for the new created port.
        xcenter: center of the component, for guessing port orientation.
        port_name_prefix: defaults to 'o' for optical and 'e' for electrical.
        port_type: optical, electrical.
        layer_label:
        fail_on_duplicates: raises ValueError for duplicated port names.
            if False adds incremental suffix (1, 2 ...) to port name.
        port_orientation: None for electrical ports.
        guess_port_orientation: assumes right: 0, left: 180, top: 90, bot: 270.
    """
    port_name_prefix_default = "o" if port_type == "optical" else "e"
    port_name_prefix = port_name_prefix or port_name_prefix_default
    yc = component.y

    port_name_to_index = {}

    xc = xcenter or component.x
    for i, label in enumerate(component.labels):
        x, y = label.origin

        if layer_label and (
            layer_label[0] != label.layer or layer_label[1] != label.texttype
        ):
            continue

        if get_name_from_label:
            port_name = label.text
        else:
            port_name = f"{port_name_prefix}{i+1}" if port_name_prefix else i

        orientation = port_orientation

        if guess_port_orientation:
            if x > xc:  # east
                orientation = 0
            elif x < xc:  # west
                orientation = 180
            elif y > yc:  # north
                orientation = 90
            elif y < yc:  # south
                orientation = 270

        if fail_on_duplicates and port_name in component.ports:
            component_ports = list(component.ports.keys())
            raise ValueError(
                f"port {port_name!r} already in {component_ports}. "
                "You can pass a port_name_prefix to add it with a different name."
            )
        if get_name_from_label and port_name in component.ports:
            port_name_to_index[label.text] = (
                port_name_to_index[label.text] + 1
                if label.text in port_name_to_index
                else 1
            )
            port_name = f"{label.text}{port_name_to_index[label.text]}"

        component.add_port(
            name=port_name,
            center=(x, y),
            width=port_width,
            orientation=orientation,
            port_type=port_type,
            layer=port_layer,
        )
    return component


def add_ports_from_siepic_pins(
    component: Component,
    pin_layer_optical: LayerSpec = "PORT",
    port_layer_optical: LayerSpec | None = None,
    pin_layer_electrical: LayerSpec = "PORTE",
    port_layer_electrical: LayerSpec | None = None,
) -> Component:
    """Add ports from SiEPIC-type cells, where the pins are defined as paths.

    Looks for label, path pairs.

    Args:
        component: component.
        pin_layer_optical: layer for optical pins.
        port_layer_optical: layer for optical ports.
        pin_layer_electrical: layer for electrical pins.
        port_layer_electrical: layer for electrical ports.
    """
    pin_layers = {"optical": pin_layer_optical, "electrical": pin_layer_electrical}

    import gdsfactory as gf

    pin_layer_optical = gf.get_layer(pin_layer_optical)
    port_layer_optical = gf.get_layer(port_layer_optical)
    pin_layer_electrical = gf.get_layer(pin_layer_electrical)
    port_layer_electrical = gf.get_layer(port_layer_electrical)

    c = component
    labels = c.get_labels()

    for path in c.paths:
        p1, p2 = path.spine()

        path_layers = list(zip(path.layers, path.datatypes))

        # Find the center of the path
        center = (p1 + p2) / 2

        # Find the label closest to the pin
        label = None
        for i, _label in enumerate(labels):
            if (
                all(isclose(_label.origin, center))
                or all(isclose(_label.origin, p1))
                or all(isclose(_label.origin, p2))
            ):
                label = _label
                labels.pop(i)
        if label is None:
            print(
                f"Warning: label not found for path: in center={center} p1={p1} p2={p2}"
            )
            continue
        if pin_layer_optical in path_layers:
            port_type = "optical"
            port_layer = port_layer_optical or None
        elif pin_layer_electrical in path_layers:
            port_type = "electrical"
            port_layer = port_layer_electrical or None
        else:
            continue

        port_name = str(label.text)

        # If the port name is already used, add a number to it
        i = 1
        while port_name in c.ports:
            port_name += f"_{i}"

        angle = round(degrees(arctan2(p2[1] - p1[1], p2[0] - p1[0])) % 360)

        port = Port(
            name=port_name,
            center=center,
            width=path.widths()[0][0],
            orientation=angle,
            layer=port_layer or pin_layers[port_type],
            port_type=port_type,
        )
        c.add_port(port)
    return c



"""Component references.

Adapted from PHIDL https://github.com/amccaugh/phidl/ by Adam McCaughan
"""

from __future__ import annotations

import typing
from typing import Any, cast

import gdstk
import numpy as np
from numpy import cos, mod, ndarray, pi, sin

from gdsfactory.component_layout import (
    Polygon,
    _GeometryHelper,
    get_polygons,
    pprint_ports,
)
from gdsfactory.port import (
    Port,
    map_ports_layer_to_orientation,
    map_ports_to_orientation_ccw,
    map_ports_to_orientation_cw,
    select_ports,
)
from gdsfactory.snap import snap_to_grid

if typing.TYPE_CHECKING:
    import shapely

    from gdsfactory.component import Component, Coordinate, Coordinates


class SizeInfo:
    def __init__(self, bbox: ndarray) -> None:
        """Initialize this object."""
        self.west = bbox[0, 0]
        self.east = bbox[1, 0]
        self.south = bbox[0, 1]
        self.north = bbox[1, 1]

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = 0.5 * (self.east + self.west)
        yc = 0.5 * (self.north + self.south)

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    def get_rect(
        self, padding=0, padding_w=None, padding_e=None, padding_n=None, padding_s=None
    ) -> tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
        w, e, s, n = self.west, self.east, self.south, self.north

        padding_n = padding if padding_n is None else padding_n
        padding_e = padding if padding_e is None else padding_e
        padding_w = padding if padding_w is None else padding_w
        padding_s = padding if padding_s is None else padding_s

        w = w - padding_w
        e = e + padding_e
        s = s - padding_s
        n = n + padding_n

        return ((w, s), (e, s), (e, n), (w, n))

    @property
    def rect(self) -> tuple[Coordinate, Coordinate]:
        return self.get_rect()

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"w: {self.west}\ne: {self.east}\ns: {self.south}\nn: {self.north}\n"


def _rotate_points(
    points: Coordinates,
    angle: float = 45.0,
    center: Coordinate = (
        0.0,
        0.0,
    ),
) -> ndarray:
    """Rotates points around a center point.

    accepts single points [1,2] or array-like[N][2], and will return in kind

    Args:
        points: rotate points around center point.
        angle: in degrees.
        center: x, y.
    """
    # First check for common, easy values of angle
    p_arr = np.asarray(points)
    if angle == 0:
        return p_arr

    c0 = np.asarray(center)
    displacement = p_arr - c0
    if angle == 180:
        return c0 - displacement

    if p_arr.ndim == 2:
        perpendicular = displacement[:, ::-1]
    elif p_arr.ndim == 1:
        perpendicular = displacement[::-1]

    # Fall back to trigonometry
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    return displacement * ca + perpendicular * sa + c0


class ComponentReference(_GeometryHelper):
    """Pointer to a Component with x, y, rotation, mirror.

    Args:
        component: Component The referenced Component.
        columns: Number of columns in the array.
        rows: Number of rows in the array.
        spacing: Distances between adjacent columns and adjacent rows.
        origin: array-like[2] of int or float
            Position where the cell is inserted.
        rotation : int or float
            Angle of rotation of the reference (in `degrees`).
        magnification : int or float
            Magnification factor for the reference.
        x_reflection : bool
            If True, the reference is reflected parallel to the x direction
            before being rotated.
        name : str (optional)
            A name for the reference (if provided).

    """

    def __init__(
        self,
        component: Component,
        origin: Coordinate = (0, 0),
        rotation: float = 0,
        magnification: float = 1,
        x_reflection: bool = False,
        visual_label: str = "",
        columns: int = 1,
        rows: int = 1,
        spacing=None,
        name: str | None = None,
        v1: tuple[float, float] | None = None,
        v2: tuple[float, float] | None = None,
    ) -> None:
        """Initialize the ComponentReference object."""
        self._reference = gdstk.Reference(
            cell=component._cell,
            origin=origin,
            rotation=np.deg2rad(rotation),
            magnification=magnification,
            x_reflection=x_reflection,
            columns=columns,
            rows=rows,
            spacing=spacing,
        )
        if v1 or v2:
            self._reference.repetition = gdstk.Repetition(
                columns=columns, rows=rows, v1=v1, v2=v2
            )

        self._ref_cell = component
        self._owner = None
        self._name = name

        # The ports of a ComponentReference have their own unique id (uid),
        # since two ComponentReferences of the same parent Component can be
        # in different locations and thus do not represent the same port
        self._local_ports = {
            name: port._copy() for name, port in component.ports.items()
        }
        self.visual_label = visual_label
        # self.uid = str(uuid.uuid4())[:8]

    @property
    def v1(self) -> tuple[float, float] | None:
        return self._reference.repetition.v1

    @property
    def v2(self) -> tuple[float, float] | None:
        return self._reference.repetition.v2

    @property
    def rows(self) -> int:
        return self._reference.repetition.rows or 1

    @property
    def columns(self) -> int:
        return self._reference.repetition.columns or 1

    @property
    def spacing(self) -> tuple[float, float] | None:
        return self._reference.repetition.spacing

    @property
    def ref_cell(self):
        return self._ref_cell

    @property
    def parent(self):
        return self._ref_cell

    @property
    def origin(self):
        return self._reference.origin

    @origin.setter
    def origin(self, value) -> None:
        self._reference.origin = value

    @property
    def magnification(self) -> float:
        return self._reference.magnification

    @magnification.setter
    def magnification(self, value) -> None:
        self._reference.magnification = value

    @property
    def rotation(self) -> float:
        return np.rad2deg(self._reference.rotation)

    @rotation.setter
    def rotation(self, value) -> None:
        self._reference.rotation = np.deg2rad(value)

    @property
    def x_reflection(self) -> bool:
        return self._reference.x_reflection

    @x_reflection.setter
    def x_reflection(self, value) -> None:
        self._reference.x_reflection = value

    def _set_ref_cell(self, value) -> None:
        self._ref_cell = value
        self._reference.cell = value._cell

    @ref_cell.setter
    def ref_cell(self, value) -> None:
        self._set_ref_cell(value)

    @parent.setter
    def parent(self, value) -> None:
        self._set_ref_cell(value)

    def get_polygon_enclosure(self) -> shapely.Polygon:
        import shapely

        return shapely.Polygon(self._reference.convex_hull())

    def get_polygon_bbox(
        self,
        default: float = 0.0,
        top: float | None = None,
        bottom: float | None = None,
        right: float | None = None,
        left: float | None = None,
    ) -> shapely.Polygon:
        """Returns shapely Polygon with padding.

        Args:
            default: default padding in um.
            top: north padding in um.
            bottom: south padding in um.
            right: east padding in um.
            left: west padding in um.
        """
        import shapely

        (xmin, ymin), (xmax, ymax) = self.bbox
        top = top if top is not None else default
        bottom = bottom if bottom is not None else default
        right = right if right is not None else default
        left = left if left is not None else default
        points = [
            [xmin - left, ymin - bottom],
            [xmax + right, ymin - bottom],
            [xmax + right, ymax + top],
            [xmin - left, ymax + top],
        ]
        return shapely.Polygon(points)

    def get_polygons(
        self,
        by_spec: bool = False,
        depth: int | None = None,
        include_paths: bool = True,
        as_array: bool = True,
        as_shapely: bool = False,
        as_shapely_merged: bool = False,
    ) -> list[Polygon] | dict[tuple[int, int], list[Polygon]]:
        """Return the list of polygons created by this reference.

        Args:
            by_spec : bool or tuple
                If True, the return value is a dictionary with the
                polygons of each individual pair (layer, datatype).
                If set to a tuple of (layer, datatype), only polygons
                with that specification are returned.
            depth : integer or None
                If not None, defines from how many reference levels to
                retrieve polygons.  References below this level will result
                in a bounding box.  If `by_spec` is True the key will be the
                name of the referenced cell.
            include_paths: If True, polygonal representation of paths are also included in the result.
            as_array: when as_array=false, return the Polygon objects instead.
                polygon objects have more information (especially when by_spec=False)
                and are faster to retrieve.
            as_shapely: returns shapely polygons.
            as_shapely_merged: returns a shapely polygonize.

        Returns
            out : list of array-like[N][2] or dictionary
                List containing the coordinates of the vertices of each
                polygon, or dictionary with the list of polygons (if
                `by_spec` is True).

        Note:
            Instances of `FlexPath` and `RobustPath` are also included in
            the result by computing their polygonal boundary.
        """
        return get_polygons(
            instance=self,
            by_spec=by_spec,
            depth=depth,
            include_paths=include_paths,
            as_array=as_array,
            as_shapely=as_shapely,
            as_shapely_merged=as_shapely_merged,
        )

    def get_labels(self, depth: int | None = None, set_transform: bool = True):
        """Return the list of labels created by this reference.

        Args:
            depth: If not None, defines from how many reference levels to retrieve labels from.
            set_transform: If True, labels will include the transformations from the reference.

        Returns:
            List containing the labels in this cell and its references.
        """
        if set_transform:
            return self._reference.get_labels(depth=depth)
        else:
            return self.parent.get_labels(depth=depth)

    def get_bounding_box(self):
        return self._reference.bounding_box()

    @property
    def layers(self) -> set[tuple[int, int]]:
        return self.parent.layers

    @property
    def settings(self):
        return self.parent.settings

    def get_paths(self, depth: int | None = None):
        """Return the list of paths created by this reference.

        Args:
            depth: If not None, defines from how many reference levels to retrieve paths from.

        Returns:
            List containing the paths in this cell and its references.
        """
        return self._reference.get_paths(depth=depth)

    def translate(self, dx, dy) -> ComponentReference:
        x0, y0 = self._reference.origin
        self.origin = (x0 + dx, y0 + dy)
        return self

    def area(self, by_spec: bool = False):
        """Calculate total area.

        Args:
            by_spec: If True, the return value is a dictionary with the areas of each individual pair (layer, datatype).

        Returns:
            out: number or dictionary Area of this cell.
        """
        return self._reference.area(by_spec=by_spec)

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if self.owner is None or value is None:
            self._owner = value
        elif value != self._owner:
            raise ValueError(
                f"Cannot reset owner of a reference once it has already been set!"
                f" Reference: {self}. Current owner: {self._owner}. "
                f"Attempting to re-assign to {value!r}"
            )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if value != self._name:
            if self.owner and value in self.owner.named_references:
                raise ValueError(
                    f"This reference's owner already has a reference with name {value!r}. Please choose another name."
                )
            if self.owner:
                self.owner._named_references.pop(self._name, None)
                self.owner._named_references[value] = self
            self._name = value

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f'ComponentReference (parent Component "{self.parent.name}", ports {list(self.ports.keys())}, origin {self.origin}, rotation {self.rotation}, x_reflection {self.x_reflection})'

    def to_dict(self):
        d = self.parent.to_dict()
        d.update(
            origin=self.origin,
            rotation=self.rotation,
            magnification=self.magnification,
            x_reflection=self.x_reflection,
        )
        return d

    @property
    def bbox(self):
        """Return the bounding box of the ComponentReference.

        it snaps to 3 decimals in um (0.001um = 1nm precision)
        """
        bbox = self.get_bounding_box()
        if bbox is None:
            bbox = ((0, 0), (0, 0))
        return snap_to_grid(np.array(bbox))

    @classmethod
    def __get_validators__(cls):
        """Get validators."""
        yield cls.validate

    @classmethod
    def validate(cls, v, _info):
        """Check with pydantic ComponentReference valid type."""
        assert isinstance(
            v, ComponentReference
        ), f"TypeError, Got {type(v)}, expecting ComponentReference"
        return v

    def __getitem__(self, key):
        """Access reference ports."""
        if key not in self.ports:
            ports = list(self.ports.keys())
            raise ValueError(f"{key!r} not in {ports}")

        return self.ports[key]

    @property
    def ports(self) -> dict[str, Port]:
        """This property allows you to access myref.ports, and receive a copy.

        of the ports dict which is correctly rotated and translated.
        """
        for name, port in self.parent.ports.items():
            port = self.parent.ports[name]
            new_center, new_orientation = self._transform_port(
                port.center,
                port.orientation,
                self.origin,
                self.rotation,
                self.x_reflection,
            )
            if name not in self._local_ports:
                self._local_ports[name] = port.copy()
            self._local_ports[name].center = new_center
            self._local_ports[name].orientation = (
                mod(new_orientation, 360) if new_orientation else new_orientation
            )
            self._local_ports[name].parent = self
        # Remove any ports that no longer exist in the reference's parent
        parent_names = self.parent.ports.keys()
        local_names = list(self._local_ports.keys())
        for name in local_names:
            if name not in parent_names:
                self._local_ports.pop(name)
        for k in list(self._local_ports):
            self._local_ports[k].reference = self
        return self._local_ports

    @property
    def info(self) -> dict[str, Any]:
        return self.parent.info

    @property
    def metadata_child(self) -> dict[str, Any]:
        return self.parent.metadata_child

    @property
    def size_info(self) -> SizeInfo:
        return SizeInfo(self.bbox)

    def pprint_ports(self) -> None:
        """Pretty print component ports."""
        pprint_ports(self.ports)

    def _transform_port(
        self,
        point: ndarray,
        orientation: float,
        origin: Coordinate = (0, 0),
        rotation: int | None = None,
        x_reflection: bool = False,
    ) -> tuple[ndarray, float]:
        """Apply GDS-type transformation to a port (x_ref)."""
        new_point = np.array(point)
        new_orientation = orientation

        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = None if orientation is None else -orientation
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
            if orientation is not None:
                new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)

        if orientation is not None:
            new_orientation = mod(new_orientation, 360)

        return new_point, new_orientation

    def _transform_point(
        self,
        point: ndarray,
        origin: Coordinate = (0, 0),
        rotation: int | None = None,
        x_reflection: bool = False,
    ) -> ndarray:
        """Apply GDS-type transformation to a point."""
        new_point = np.array(point)

        if x_reflection:
            new_point[1] = -new_point[1]
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
        if origin is not None:
            new_point = new_point + np.array(origin)

        return new_point

    def move(
        self,
        origin: Port | Coordinate | str = (0, 0),
        destination: Port | Coordinate | str | None = None,
        axis: str | None = None,
    ) -> ComponentReference:
        """Move the ComponentReference from origin point to destination.

        Both origin and destination can be 1x2 array-like, Port, or a key
        corresponding to one of the Ports in this device_ref.

        Args:
            origin: Port, port_name or Coordinate.
            destination: Port, port_name or Coordinate.
            axis: for the movement.

        Returns:
            ComponentReference.
        """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0, 0)

        if isinstance(origin, str):
            if origin not in self.ports:
                raise ValueError(f"{origin} not in {self.ports.keys()}")

            origin = self.ports[origin]
            origin = cast(Port, origin)
            o = origin.center
        elif hasattr(origin, "center"):
            origin = cast(Port, origin)
            o = origin.center
        elif np.array(origin).size == 2:
            o = origin
        else:
            raise ValueError(
                f"move(origin={origin})\n"
                f"Invalid origin = {origin!r} needs to be"
                f"a coordinate, port or port name {list(self.ports.keys())}"
            )

        if isinstance(destination, str):
            if destination not in self.ports:
                raise ValueError(f"{destination} not in {self.ports.keys()}")

            destination = self.ports[destination]
            destination = cast(Port, destination)
            d = destination.center
        if hasattr(destination, "center"):
            destination = cast(Port, destination)
            d = destination.center
        elif np.array(destination).size == 2:
            d = destination

        else:
            raise ValueError(
                f"{self.parent.name}.move(destination={destination!r}) \n"
                f"Invalid destination = {destination!r} needs to be"
                f"a coordinate, a port, or a valid port name {list(self.ports.keys())}"
            )

        # Lock one axis if necessary
        if axis == "x":
            d = (d[0], o[1])
        if axis == "y":
            d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        self._bb_valid = False
        return self

    def rotate(
        self,
        angle: float = 45,
        center: Coordinate | str | int = (0.0, 0.0),
    ) -> ComponentReference:
        """Return rotated ComponentReference.

        Args:
            angle: in degrees.
            center: x, y.
        """
        if angle == 0:
            return self
        if isinstance(center, int | str):
            center = self.ports[center].center

        if isinstance(center, Port):
            center = center.center
        self.rotation += angle
        self.rotation %= 360
        self.origin = _rotate_points(self.origin, angle, center)
        self._bb_valid = False
        return self

    def mirror_x(
        self, port_name: str | None = None, x0: Coordinate | None = None
    ) -> ComponentReference:
        """Perform horizontal mirror using x0 or port as axis (default, x0=0).

        This is the default for mirror along X=x0 axis
        """
        if port_name is None and x0 is None:
            x0 = -self.x

        if port_name is not None:
            position = self.ports[port_name]
            x0 = position.x
        self.mirror((x0, 1), (x0, 0))
        return self

    def mirror_y(
        self, port_name: str | None = None, y0: float | None = None
    ) -> ComponentReference:
        """Perform vertical mirror using y0 as axis (default, y0=0)."""
        if port_name is None and y0 is None:
            y0 = 0.0

        if port_name is not None:
            position = self.ports[port_name]
            y0 = position.y
        self.mirror((1, y0), (0, y0))
        return self

    def mirror(
        self,
        p1: Coordinate = (0.0, 1.0),
        p2: Coordinate = (0.0, 0.0),
    ) -> ComponentReference:
        """Mirrors.

        Args:
            p1: point 1.
            p2: point 2.
        """
        if isinstance(p1, Port):
            p1 = p1.center
        if isinstance(p2, Port):
            p2 = p2.center
        p1 = np.array(p1)
        p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1

        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        self.origin = _rotate_points(self.origin, angle=-angle, center=(0, 0))
        self.rotation -= angle

        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin = (self.origin[0], -1 * self.origin[1])

        self.rotation = -1 * self.rotation

        # Un-rotate and un-translate
        self.origin = _rotate_points(self.origin, angle=angle, center=(0, 0))
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = self.origin + p1

        self._bb_valid = False
        return self

    def connect(
        self,
        port: str | Port,
        destination: Port,
        overlap: float = 0.0,
        preserve_orientation: bool = False,
    ) -> ComponentReference:
        """Return ComponentReference where port connects to a destination.

        Args:
            port: origin (port, or port name) to connect.
            destination: destination port.
            overlap: how deep does the port go inside.
            preserve_orientation: True, does not rotate the reference to align port
                orientation and reference keep its orientation pre-connection.

        Returns:
            ComponentReference: with correct rotation to connect to destination.
        """
        # port can either be a string with the name, port index, or an actual Port
        if port in self.ports:
            p = self.ports[port]
        elif isinstance(port, Port):
            p = port
        else:
            ports = list(self.ports.keys())
            raise ValueError(
                f"port = {port!r} not in {self.parent.name!r} ports {ports}"
            )

        if (
            destination.orientation is not None
            and p.orientation is not None
            and not preserve_orientation
        ):
            angle = 180 + destination.orientation - p.orientation
            angle = angle % 360
            self.rotate(angle=angle, center=p.center)

        self.move(origin=p, destination=destination)

        if destination.orientation is not None:
            self.move(
                -overlap
                * np.array(
                    [
                        cos(destination.orientation * pi / 180),
                        sin(destination.orientation * pi / 180),
                    ]
                )
            )

        return self

    def get_ports_list(self, **kwargs) -> list[Port]:
        """Return a list of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return list(select_ports(self.ports, **kwargs).values())

    def get_ports_dict(self, **kwargs) -> dict[str, Port]:
        """Return a dict of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return select_ports(self.ports, **kwargs)

    @property
    def ports_layer(self) -> dict[str, str]:
        """Return a mapping from layer0_layer1_E0: portName."""
        return map_ports_layer_to_orientation(self.ports)

    def port_by_orientation_cw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_cw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def port_by_orientation_ccw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_ccw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def get_ports_xsize(self, **kwargs) -> float:
        """Return xdistance from east to west ports.

        Keyword Args:
            kwargs: orientation, port_type, layer.
        """
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return ports_ccw[0].x - ports_cw[0].x

    def get_ports_ysize(self, **kwargs) -> float:
        """Returns ydistance from east to west ports."""
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return ports_ccw[0].y - ports_cw[0].y


if __name__ == "__main__":
    # test_get_polygons_ref()
    # test_get_polygons()
    import gdsfactory as gf

    c = gf.Component("parent")
    ref = c << gf.components.straight()
    c.add_ports(ref.ports)
    ref.movex(5)
    # assert c.ports['o1'].center[0] == 5, print(c.ports['o1'])
    print(c.ports["o1"].center)
    c.show(show_ports=True)

    # p = ref.get_polygons(by_spec=(1, 0), as_array=False)

    # c = gf.Component("parent")
    # c2 = gf.Component("child")
    # length = 10
    # width = 0.5
    # layer = (1, 0)
    # c2.add_polygon([(0, 0), (length, 0), (length, width), (0, width)], layer=layer)
    # c << c2

    # c = gf.c.dbr()
    # c.show()

    # import gdsfactory as gf

    # c = gf.Component()
    # mzi = c.add_ref(gf.components.mzi())
    # bend = c.add_ref(gf.components.bend_euler())
    # bend.move("o1", mzi.ports["o2"])
    # bend.move("o1", "o2")



import numpy as np

import gdsfactory as gf
from gdsfactory.typings import ComponentSpec, LayerSpec


@gf.cell
def layer_priority(
    component: ComponentSpec,
    layer_high_order: LayerSpec,
    layer_low_order: LayerSpec,
    remove_high_order: bool = False,
    **kwargs,
) -> gf.Component:
    """Returns new component after removing one layer from another.

    Args:
        component: spec.
        layer_high_order: layer used to etch.
        layer_low_order: layer etched into.
        remove_high_order: whether to also remove the high order layer polygons. \
                Useful if the higher order layer is purely logical.
        kwargs: keyword arguments for boolean difference operation.
    """
    c = gf.Component()

    component = gf.get_component(component)

    # Obtain component subsets
    layers_to_remove = (
        [gf.get_layer(layer_low_order), gf.get_layer(layer_high_order)]
        if remove_high_order
        else [gf.get_layer(layer_low_order)]
    )
    component_minus_layers = component.extract(
        [
            gf.get_layer(layer)
            for layer in component.get_layers()
            if gf.get_layer(layer) not in layers_to_remove
        ]
    )
    component_high_order = component.extract([layer_high_order])
    component_low_order = component.extract([layer_low_order])

    # Remove high priority from low priority
    component_high_order_removed_from_low_order = gf.geometry.boolean(
        A=component_low_order,
        B=component_high_order,
        operation="A-B",
        layer=layer_low_order,
        **kwargs,
    )

    # Place all the other layers
    a = c << component_minus_layers
    b = c << component_high_order_removed_from_low_order

    c.absorb(a)
    c.absorb(b)
    c.add_ports(ports=component.ports)
    c.copy_child_info(component)
    return c


def test_layer_priority() -> None:
    funky_cross_section = gf.cross_section.cross_section(
        width=0.5,
        layer="WG",
        sections=[
            gf.cross_section.Section(width=2, layer="N"),
            gf.cross_section.Section(width=2, layer="SLAB150"),
        ],
    )

    c = gf.components.coupler(cross_section=funky_cross_section)

    c_WG_before = c.extract([gf.get_layer("WG")])
    c_SLAB150_before = c.extract([gf.get_layer("SLAB150")])
    c_N_before = c.extract([gf.get_layer("N")])

    c_processed = layer_priority(
        component=c,
        layer_high_order="WG",
        layer_low_order="SLAB150",
        remove_high_order=False,
    )

    c_WG_after = c_processed.extract([gf.get_layer("WG")])
    c_SLAB150_after = c_processed.extract([gf.get_layer("SLAB150")])
    c_N_after = c_processed.extract([gf.get_layer("N")])

    assert np.isclose(c_WG_before.area(), c_WG_after.area(), atol=1e-3)
    assert np.isclose(c_N_before.area(), c_N_after.area(), atol=1e-3)
    assert c_SLAB150_before.area() > c_SLAB150_after.area()

    c_processed = layer_priority(
        component=c,
        layer_high_order="WG",
        layer_low_order="SLAB150",
        remove_high_order=True,
    )
    c_WG_after = c_processed.extract([gf.get_layer("WG")])

    assert c_WG_after.area() == 0


if __name__ == "__main__":
    funky_cross_section = gf.cross_section.cross_section(
        width=0.5,
        layer="WG",
        sections=[
            gf.cross_section.Section(width=2, layer="N"),
            gf.cross_section.Section(width=2, layer="SLAB150"),
        ],
    )

    c = gf.components.coupler(cross_section=funky_cross_section)

    c2 = layer_priority(
        component=c,
        layer_high_order="WG",
        layer_low_order="SLAB150",
        remove_high_order=False,
    )
    c2.show(show_ports=True)



from __future__ import annotations

from collections.abc import Callable

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.pad import pad_rectangular
from gdsfactory.components.straight_heater_metal import straight_heater_metal
from gdsfactory.port import select_ports_electrical
from gdsfactory.routing.route_fiber_array import route_fiber_array
from gdsfactory.routing.sort_ports import sort_ports_x
from gdsfactory.typings import (
    ComponentSpec,
    CrossSectionSpec,
    LayerSpec,
    Strs,
)


@cell
def add_pads_bot(
    component: ComponentSpec = straight_heater_metal,
    select_ports: Callable = select_ports_electrical,
    port_names: Strs | None = None,
    component_name: str | None = None,
    cross_section: CrossSectionSpec = "xs_metal_routing",
    get_input_labels_function: Callable | None = None,
    layer_label: LayerSpec = "TEXT",
    pad_port_name: str = "e1",
    pad_port_labels: tuple[str, ...] | None = None,
    pad: ComponentSpec = pad_rectangular,
    bend: ComponentSpec = "wire_corner",
    straight_separation: float | None = None,
    pad_spacing: float | str = "pad_spacing",
    optical_routing_type: int | None = 1,
    **kwargs,
) -> Component:
    """Returns new component with ports connected bottom pads.

    Args:
        component: component spec to connect to.
        select_ports: function to select_ports.
        port_names: optional port names. Overrides select_ports.
        component_name: optional for the label.
        cross_section: cross_section spec.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        pad_port_name: pad input port name.
        pad_port_labels: pad list of labels.
        pad: spec for route terminations.
        bend: bend spec.
        straight_separation: from wire edge to edge. Defaults to xs.width+xs.gap
        pad_spacing: in um. Defaults to pad_spacing constant from the PDK.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.

    Keyword Args:
        straight: straight spec.
        taper: taper spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        fanout_length: if None, automatic calculation of fanout length.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: get_route.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.c.nxn(
            xsize=600,
            ysize=200,
            north=2,
            south=3,
            wg_width=10,
            layer="M3",
            port_type="electrical",
        )
        cc = gf.routing.add_pads_bot(component=c, port_names=("e1", "e4"), fanout_length=50)
        cc.plot()

    """
    component_new = Component()
    component = gf.get_component(component)
    component_name = component_name or component.name

    pad_spacing = gf.get_constant(pad_spacing)
    cref = component_new << component
    ports = [cref[port_name] for port_name in port_names] if port_names else None
    ports = ports or select_ports(cref.ports)
    xs = gf.get_cross_section(cross_section)

    straight_separation = straight_separation or xs.width + xs.gap

    pad_component = gf.get_component(pad)
    if pad_port_name not in pad_component.ports:
        pad_ports = list(pad_component.ports.keys())
        raise ValueError(
            f"pad_port_name = {pad_port_name!r} not in {pad_component.name!r} ports {pad_ports}"
        )

    pad_orientation = int(pad_component[pad_port_name].orientation)
    if pad_orientation != 180:
        raise ValueError(
            f"port.orientation={pad_orientation} for port {pad_port_name!r} needs to be 180 degrees."
        )

    if not ports:
        raise ValueError(
            f"select_ports or port_names did not match any ports in {list(component.ports.keys())}"
        )

    (
        elements,
        io_gratings_lines,
        ports_grating_input_waveguide,
        ports_loopback,
        ports_component,
    ) = route_fiber_array(
        component=component,
        grating_coupler=pad,
        gc_port_name=pad_port_name,
        component_name=component_name,
        cross_section=cross_section,
        select_ports=select_ports,
        get_input_labels_function=get_input_labels_function,
        layer_label=layer_label,
        with_loopback=False,
        bend=bend,
        straight_separation=straight_separation,
        port_names=port_names,
        fiber_spacing=pad_spacing,
        optical_routing_type=optical_routing_type,
        **kwargs,
    )
    if len(elements) == 0:
        return component

    for e in elements:
        component_new.add(e)
    for io_gratings in io_gratings_lines:
        component_new.add(io_gratings)

    component_new.add_ref(component)

    for port in component.ports.values():
        if port not in ports:
            component_new.add_port(port.name, port=port)

    ports = sort_ports_x(ports_grating_input_waveguide + ports_loopback)

    if pad_port_labels:
        for gc_port_label, port in zip(pad_port_labels, ports):
            if layer_label:
                component_new.add_label(
                    text=gc_port_label, layer=layer_label, position=port.center
                )

    for port_component, port_grating in zip(
        ports_component, ports_grating_input_waveguide
    ):
        grating_ref = port_grating.parent
        component_new.add_port(
            f"elec-{grating_ref.parent.name}-{component_name}-{port_component.name}",
            port=port_grating,
        )

    for i, port in enumerate(ports_loopback):
        grating_ref = port_grating.parent
        component_new.add_port(
            f"elec-{grating_ref.parent.name}-{component_name}-loopback{i}",
            port=port,
        )

    component_new.copy_child_info(component)
    return component_new


@gf.cell
def add_pads_top(
    component: ComponentSpec = straight_heater_metal, **kwargs
) -> Component:
    """Returns new component with ports connected top pads.

    Args:
        component: component spec to connect to.

    Keyword Args:
        select_ports: function to select_ports.
        port_names: optional port names. Overrides select_ports.
        component_name: optional for the label.
        cross_section: cross_section function.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        pad_port_name: pad input port name.
        pad_port_labels: pad list of labels.
        pad: spec for route terminations.
        bend: bend spec.
        straight_separation: from edge to edge.
        straight: straight spec.
        taper: taper spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        fanout_length: if None, automatic calculation of fanout length.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: get_route.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.

    .. plot::
        :include-source:

        import gdsfactory as gf
        c = gf.c.nxn(
            xsize=600,
            ysize=200,
            north=2,
            south=3,
            wg_width=10,
            layer="M3",
            port_type="electrical",
        )
        cc = gf.routing.add_pads_top(component=c, port_names=("e1", "e4"), fanout_length=50)
        cc.plot()

    """
    c = Component()
    _c = add_pads_bot(component=component, **kwargs)
    ref = c << _c
    ref.mirror_y()
    c.add_ports(ref.ports)
    c.copy_child_info(_c)
    return c


if __name__ == "__main__":
    # c = gf.components.pad()
    # c = gf.components.straight_heater_metal(length=100.0)
    # c = gf.components.straight(length=100.0)

    # cc = add_pads_top(component=c, port_names=("e1",))
    # cc = add_pads_top(component=c, port_names=("e1", "e2"), fanout_length=50)
    c = gf.c.nxn(
        xsize=600,
        ysize=200,
        north=2,
        south=3,
        wg_width=10,
        layer="M3",
        port_type="electrical",
    )
    cc = add_pads_top(component=c)
    cc.show(show_ports=True)



from __future__ import annotations

from functools import partial

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.typings import ComponentSpec, CrossSectionSpec


@cell
def straight_heater_metal_undercut(
    length: float = 320.0,
    length_undercut_spacing: float = 6.0,
    length_undercut: float = 30.0,
    length_straight: float = 0.1,
    length_straight_input: float = 15.0,
    cross_section: CrossSectionSpec = "xs_sc",
    cross_section_heater: CrossSectionSpec = "xs_heater_metal",
    cross_section_waveguide_heater: CrossSectionSpec = "xs_sc_heater_metal",
    cross_section_heater_undercut: CrossSectionSpec = "xs_sc_heater_metal_undercut",
    with_undercut: bool = True,
    via_stack: ComponentSpec | None = "via_stack_heater_mtop",
    port_orientation1: int | None = None,
    port_orientation2: int | None = None,
    heater_taper_length: float | None = 5.0,
    ohms_per_square: float | None = None,
) -> Component:
    """Returns a thermal phase shifter.

    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide.
        length_undercut_spacing: from undercut regions.
        length_undercut: length of each undercut section.
        length_straight_input: from input port to where trenches start.
        cross_section_heater: for heated sections. heater metal only.
        cross_section_waveguide_heater: for heated sections.
        cross_section_heater_undercut: for heated sections with undercut.
        with_undercut: isolation trenches for higher efficiency.
        via_stack: via stack.
        port_orientation1: left via stack port orientation.
        port_orientation2: right via stack port orientation.
        heater_taper_length: minimizes current concentrations from heater to via_stack.
        ohms_per_square: to calculate resistance.
        cross_section: for waveguide ports.
    """
    xs = gf.get_cross_section(cross_section)
    period = length_undercut + length_undercut_spacing
    n = int((length - 2 * length_straight_input) // period)

    length_straight_input = (length - n * period) / 2

    if length_straight > length_straight_input:
        raise ValueError("length_straight_ must be smaller than length_straight_input")

    length_straight_input -= length_straight

    s_ports = gf.components.straight(
        cross_section=cross_section, length=length_straight, add_pins=False
    )

    s_si = gf.components.straight(
        cross_section=cross_section_waveguide_heater,
        length=length_straight_input,
        add_pins=False,
    )
    cross_section_undercut = (
        cross_section_heater_undercut
        if with_undercut
        else cross_section_waveguide_heater
    )
    s_uc = gf.components.straight(
        cross_section=cross_section_undercut, length=length_undercut, add_pins=False
    )
    s_spacing = gf.components.straight(
        cross_section=cross_section_waveguide_heater,
        length=length_undercut_spacing,
        add_pins=False,
    )
    symbol_to_component = {
        "_": (s_ports, "o1", "o2"),
        "-": (s_si, "o1", "o2"),
        "U": (s_uc, "o1", "o2"),
        "H": (s_spacing, "o1", "o2"),
    }

    # Each character in the sequence represents a component
    sequence = "_-" + n * "UH" + "-_"

    c = Component()
    sequence = gf.components.component_sequence(
        sequence=sequence, symbol_to_component=symbol_to_component
    )
    c.add_ref(sequence)
    c.add_ports(sequence.ports)

    x = gf.get_cross_section(cross_section_heater)
    heater_width = x.width
    xs.add_pins(c)

    if via_stack:
        refs = list(sequence.named_references.keys())
        via = via_stackw = via_stacke = gf.get_component(via_stack)
        via_stack_west_center = sequence.named_references[refs[1]].size_info.cw
        via_stack_east_center = sequence.named_references[refs[-2]].size_info.ce
        dx = via_stackw.get_ports_xsize() / 2 + heater_taper_length or 0

        via_stack_west = c << via_stackw
        via_stack_east = c << via_stacke
        via_stack_west.move(via_stack_west_center - (dx, 0))
        via_stack_east.move(via_stack_east_center + (dx, 0))

        valid_orientations = {p.orientation for p in via.ports.values()}
        p1 = via_stack_west.get_ports_list(orientation=port_orientation1)
        p2 = via_stack_east.get_ports_list(orientation=port_orientation2)

        if not p1:
            raise ValueError(
                f"No ports for port_orientation1 {port_orientation1} in {valid_orientations}"
            )
        if not p2:
            raise ValueError(
                f"No ports for port_orientation2 {port_orientation2} in {valid_orientations}"
            )

        c.add_ports(p1, prefix="l_")
        c.add_ports(p2, prefix="r_")
        if heater_taper_length:
            taper = gf.components.taper(
                width1=via_stackw.ports["e1"].width,
                width2=heater_width,
                length=heater_taper_length,
                cross_section=x,
            )
            taper1 = c << taper
            taper2 = c << taper
            taper1.connect("o1", via_stack_west.ports["e3"])
            taper2.connect("o1", via_stack_east.ports["e1"])

    c.info["resistance"] = (
        ohms_per_square * heater_width * length if ohms_per_square else None
    )
    return c


@cell
def straight_heater_metal_simple(
    length: float = 320.0,
    cross_section_heater: CrossSectionSpec = "xs_heater_metal",
    cross_section_waveguide_heater: CrossSectionSpec = "xs_sc_heater_metal",
    via_stack: ComponentSpec | None = "via_stack_heater_mtop",
    port_orientation1: int | None = None,
    port_orientation2: int | None = None,
    heater_taper_length: float | None = 5.0,
    ohms_per_square: float | None = None,
) -> Component:
    """Returns a thermal phase shifter that has properly fixed electrical connectivity to extract a suitable electrical netlist and models.
    dimensions from https://doi.org/10.1364/OE.27.010456

    Args:
        length: of the waveguide.
        length_undercut_spacing: from undercut regions.
        length_undercut: length of each undercut section.
        cross_section_heater: for heated sections. heater metal only.
        cross_section_waveguide_heater: for heated sections.
        cross_section_heater_undercut: for heated sections with undercut.
        with_undercut: isolation trenches for higher efficiency.
        via_stack: via stack.
        port_orientation1: left via stack port orientation.
        port_orientation2: right via stack port orientation.
        heater_taper_length: minimizes current concentrations from heater to via_stack.
        ohms_per_square: to calculate resistance.
        cross_section: for waveguide ports.
    """
    c = Component()
    straight_heater_section = gf.components.straight(
        cross_section=cross_section_waveguide_heater,
        length=length,
    )

    c.add_ref(straight_heater_section)
    c.add_ports(straight_heater_section.ports)
    x = gf.get_cross_section(cross_section_heater)
    heater_width = x.width

    if via_stack:
        via = via_stackw = via_stacke = gf.get_component(via_stack)
        via_stack_west_center = straight_heater_section.size_info.cw
        via_stack_east_center = straight_heater_section.size_info.ce
        dx = via_stackw.get_ports_xsize() / 2 + heater_taper_length or 0

        via_stack_west = c << via_stackw
        via_stack_east = c << via_stacke
        via_stack_west.move(via_stack_west_center - (dx, 0))
        via_stack_east.move(via_stack_east_center + (dx, 0))

        valid_orientations = {p.orientation for p in via.ports.values()}
        p1 = via_stack_west.get_ports_list(orientation=port_orientation1)
        p2 = via_stack_east.get_ports_list(orientation=port_orientation2)

        if not p1:
            raise ValueError(
                f"No ports for port_orientation1 {port_orientation1} in {valid_orientations}"
            )
        if not p2:
            raise ValueError(
                f"No ports for port_orientation2 {port_orientation2} in {valid_orientations}"
            )

        # c.add_ports(p1, prefix="l_")
        # c.add_ports(p2, prefix="r_")
        if heater_taper_length:
            taper = gf.components.taper(
                width1=via_stackw.ports["e1"].width,
                width2=heater_width,
                length=heater_taper_length,
                cross_section=x,
                port_order_name=("e1", "e2"),
                port_order_types=("electrical", "electrical"),
            )
            taper1 = c << taper
            taper2 = c << taper
            taper1.connect("e1", via_stack_west.ports["e3"])
            taper2.connect("e1", via_stack_east.ports["e1"])

    c.info["resistance"] = (
        ohms_per_square * heater_width * length if ohms_per_square else None
    )
    return c


straight_heater_metal = partial(
    straight_heater_metal_undercut,
    with_undercut=False,
)
straight_heater_metal_90_90 = partial(
    straight_heater_metal_undercut,
    with_undercut=False,
    port_orientation1=90,
    port_orientation2=90,
)
straight_heater_metal_undercut_90_90 = partial(
    straight_heater_metal_undercut,
    with_undercut=False,
    port_orientation1=90,
    port_orientation2=90,
)


def test_ports() -> None:
    c = straight_heater_metal(length=50.0)
    assert c.ports["o2"].center[0] == 50.0, c.ports["o2"].center[0]


if __name__ == "__main__":
    # c = straight_heater_metal_undercut()
    # print(c.ports['o2'].center[0])
    # c.pprint_ports()
    # c = straight_heater_metal(heater_width=5, length=50.0)

    c = straight_heater_metal_undercut(length=200)
    # n = c.get_netlist()
    # c = straight_heater_metal(length=20)
    c.show(show_ports=False)
    # scene = c.to_3d()
    # scene.show()



from __future__ import annotations

from collections.abc import Callable

import numpy as np

import gdsfactory as gf
from gdsfactory.path import transition_adiabatic
from gdsfactory.typings import CrossSectionSpec

adiabatic_polyfit_TE1550SOI_220nm = np.array(
    [
        1.02478963e-09,
        -8.65556534e-08,
        3.32415694e-06,
        -7.68408985e-05,
        1.19282177e-03,
        -1.31366332e-02,
        1.05721429e-01,
        -6.31057637e-01,
        2.80689677e00,
        -9.26867694e00,
        2.24535191e01,
        -3.90664800e01,
        4.71899278e01,
        -3.74726005e01,
        1.77381560e01,
        -1.12666286e00,
    ]
)


@gf.cell
def taper_adiabatic(
    width1: float = 0.5,
    width2: float = 5.0,
    length: float = 0,
    neff_w: Callable = lambda w: np.poly1d(adiabatic_polyfit_TE1550SOI_220nm)(w),
    alpha: float = 1,
    wavelength: float = 1.55,
    npoints: int = 200,
    cross_section: CrossSectionSpec = "xs_sc",
) -> gf.Component:
    """Returns a straight adiabatic_taper from an effective index callable.

    Args:
        width1: initial width.
        width2: final width.
        length: 0 uses the optimized length, and otherwise the optimal shape is compressed/stretched to the specified length.
        neff_y: a callable that returns the effective index as a function of width
                - By default, will use a compact model of neff(y) for fundamental 1550 nm TE mode of 220nm-thick core with 3.45 index, fully clad with 1.44 index. Many coefficients are needed to capture the behaviour.
        alpha: parameter that scales the rate of width change.
                - closer to 0 means longer and more adiabatic;
                - 1 is the intuitive limit beyond which higher order modes are excited;
                - [2] reports good performance up to 1.4 for fundamental TE in SOI (for multiple core thicknesses)

    References:
        [1] Burns, W. K., et al. "Optical waveguide parabolic coupling horns." Appl. Phys. Lett., vol. 30, no. 1, 1 Jan. 1977, pp. 28-30, doi:10.1063/1.89199.
        [2] Fu, Yunfei, et al. "Efficient adiabatic silicon-on-insulator waveguide taper." Photonics Res., vol. 2, no. 3, 1 June 2014, pp. A41-A44, doi:10.1364/PRJ.2.000A41.
        npoints: number of points for sampling
    """
    xs = gf.get_cross_section(cross_section)
    layer = xs.layer

    # Obtain optimal curve
    x_opt, w_opt = transition_adiabatic(
        width1, width2, neff_w=neff_w, wavelength=wavelength, alpha=alpha
    )

    # Resample the points
    from scipy import interpolate

    w_opt_interp = interpolate.interp1d(x_opt, w_opt)

    if not length:
        length = x_opt[-1]
    x = np.linspace(0, length, npoints)
    w = w_opt_interp(x)

    # Stretch/compress x
    x_array = np.linspace(0, length, npoints) * (1 + length - x_opt[-1])
    y_array = w / 2

    c = gf.Component()
    c.add_polygon(
        list(zip(x_array, y_array)) + list(zip(x_array, -y_array))[::-1], layer=layer
    )

    # Define ports
    xs1 = xs.copy(width=width1)
    xs2 = xs.copy(width=width2)

    c.add_port(
        name="o1",
        center=(0, 0),
        width=width1,
        orientation=180,
        cross_section=xs1,
        layer=layer,
    )
    c.add_port(
        name="o2",
        center=(length, 0),
        width=width2,
        orientation=0,
        cross_section=xs2,
        layer=layer,
    )
    return c


if __name__ == "__main__":
    c = taper_adiabatic(width1=0.5, width2=5)
    c.show(show_ports=True)



from __future__ import annotations

import inspect
from collections.abc import Sequence

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.interdigital_capacitor import interdigital_capacitor
from gdsfactory.typings import LayerSpec

INTERDIGITAL_DEFAULTS = {
    k: v.default
    for k, v in inspect.signature(interdigital_capacitor).parameters.items()
}


@gf.cell
def interdigital_capacitor_enclosed(
    enclosure_box: Sequence[Sequence[float | int]] = [[-200, -200], [200, 200]],
    fingers: int = INTERDIGITAL_DEFAULTS["fingers"],
    finger_length: float | int = INTERDIGITAL_DEFAULTS["finger_length"],
    finger_gap: float | int = INTERDIGITAL_DEFAULTS["finger_gap"],
    thickness: float | int = INTERDIGITAL_DEFAULTS["thickness"],
    cpw_dimensions: tuple[float | int, float | int] = (10, 6),
    gap_to_ground: float | int = 5,
    metal_layer: LayerSpec = INTERDIGITAL_DEFAULTS["layer"],
    gap_layer: LayerSpec = "DEEPTRENCH",
) -> Component:
    """Generates an interdigital capacitor surrounded by a ground plane and \
            coplanar waveguides with ports on both ends. \
            See for :func:`~interdigital_capacitor` for details.

    Note:
        ``finger_length=0`` effectively provides a plate capacitor.

    Args:
        enclosure_box: Bounding box dimensions for a ground metal enclosure.
        fingers: total fingers of the capacitor.
        finger_length: length of the probing fingers.
        finger_gap: length of gap between the fingers.
        thickness: Thickness of fingers and section before the fingers.
        gap_to_ground: Size of gap from capacitor to ground metal.
        cpw_dimensions: Dimensions for the trace width and gap width of connecting coplanar waveguides.
        metal_layer: layer for metalization.
        gap_layer: layer for trenching.
    """
    c = Component()
    cap = interdigital_capacitor(
        fingers, finger_length, finger_gap, thickness, metal_layer
    ).ref_center()
    c.add(cap)

    gap = Component()
    for port in cap.get_ports_list():
        port2 = port.copy()

        # TODO very bad check, should do vector according to orientation
        direction = -1 if port.orientation > 0 else 1
        port2.move((30 * direction, 0))
        port2 = port2.flip()

        cpw_a, cpw_b = cpw_dimensions
        s0 = gf.Section(
            width=cpw_a, offset=0, layer=metal_layer, port_names=("in", "out")
        )
        s1 = gf.Section(width=cpw_b, offset=(cpw_a + cpw_b) / 2, layer=gap_layer)
        s2 = gf.Section(width=cpw_b, offset=-(cpw_a + cpw_b) / 2, layer=gap_layer)
        x = gf.CrossSection(sections=[s0, s1, s2])
        route = gf.routing.get_route(
            port,
            port2,
            cross_section=x,
        )
        c.add(route.references)

        # TODO check size parameters, now they're maybe approximate
        term = c << gf.components.bbox(
            [[0, 0], [cpw_b, cpw_a + 2 * cpw_b]], layer=gap_layer
        )
        if direction < 0:
            term.movex(-cpw_b)
        term.move(
            destination=route.ports[-1].move_copy(-1 * np.array([0, cpw_a / 2 + cpw_b]))
        )

        c.add_port(route.ports[-1])
        c.auto_rename_ports()

    gap.add_polygon(cap.get_polygon_enclosure(), layer=gap_layer)
    gap = gap.offset(gap_to_ground, layer=gap_layer)
    gap = gf.geometry.boolean(A=gap, B=c, operation="A-B", layer=gap_layer)

    ground = gf.components.bbox(bbox=enclosure_box, layer=metal_layer)
    ground = gf.geometry.boolean(
        A=ground, B=[c, gap], operation="A-B", layer=metal_layer
    )
    _ = c << ground
    return c.flatten()


if __name__ == "__main__":
    c = interdigital_capacitor_enclosed()
    c.show(show_ports=True)



"""Design For Testing module includes test protocols."""

from __future__ import annotations

from gdsfactory.labels import ehva, siepic, write_labels
from gdsfactory.labels.add_label_yaml import add_label_json, add_label_yaml
from gdsfactory.labels.ehva import add_label_ehva
from gdsfactory.labels.merge_test_metadata import merge_test_metadata
from gdsfactory.labels.siepic import add_fiber_array_siepic

__all__ = [
    "add_fiber_array_siepic",
    "add_label_ehva",
    "add_label_yaml",
    "add_label_json",
    "ehva",
    "siepic",
    "write_labels",
    "merge_test_metadata",
]



import json
from typing import Any

import gdsfactory as gf
from gdsfactory.components.grating_coupler_elliptical import (
    grating_coupler_elliptical_te,
)
from gdsfactory.components.mzi_phase_shifter import mzi_phase_shifter
from gdsfactory.components.pad import pad_small
from gdsfactory.typings import ComponentSpec, CrossSectionSpec, LayerSpec


@gf.cell
def add_fiber_array_optical_south_electrical_north(
    component: ComponentSpec = mzi_phase_shifter,
    pad: ComponentSpec = pad_small,
    with_loopback: bool = True,
    pad_spacing: float = 100.0,
    fiber_spacing: float = 127.0,
    pad_gc_spacing: float = 250.0,
    electrical_port_names: list[str] | None = None,
    electrical_port_orientation: float | None = 90,
    npads: int | None = None,
    grating_coupler: ComponentSpec = grating_coupler_elliptical_te,
    xs_metal: CrossSectionSpec = "xs_metal_routing",
    layer_label: LayerSpec = "TEXT",
    measurement: str = "optical_loopback2_heater_sweep",
    measurement_settings: dict[str, Any] | None = None,
    analysis: str = "",
    analysis_settings: dict[str, Any] | None = None,
    doe: str = "",
    anchor: str = "sw",
    **kwargs,
) -> gf.Component:
    """Returns a fiber array with Optical gratings on South and Electrical pads on North.

    This a test configuration for DC pads.

    Args:
        component: component spec to add fiber and pads.
        pad: pad spec.
        with_loopback: whether to add a loopback port.
        pad_spacing: spacing between pads.
        fiber_spacing: spacing between grating couplers.
        pad_gc_spacing: spacing between pads and grating couplers.
        electrical_port_names: list of electrical port names. Defaults to all.
        electrical_port_orientation: orientation of electrical ports. Defaults to 90.
        npads: number of pads. Defaults to one per electrical_port_names.
        grating_coupler: grating coupler function.
        xs_metal: metal cross section.
        layer_label: layer for settings label.
        measurement: measurement name.
        measurement_settings: measurement settings.
        analysis: analysis name.
        analysis_settings: analysis settings.
        doe: Design of Experiment.
        anchor: anchor point for the label. Defaults to south-west "sw". \
            Valid options are: "n", "s", "e", "w", "ne", "nw", "se", "sw", "c".

    Keyword Args:
        gc_port_name: grating coupler input port name.
        gc_port_labels: grating coupler list of labels.
        component_name: optional for the label.
        select_ports: function to select ports.
        cross_section: cross_section function.
        get_input_labels_function: function to get input labels. None skips labels.
        layer_label: optional layer for grating coupler label.
        bend: bend spec.
        straight: straight spec.
        taper: taper spec.
        get_input_label_text_loopback_function: function to get input label test.
        get_input_label_text_function: for labels.
        fanout_length: if None, automatic calculation of fanout length.
        max_y0_optical: in um.
        with_loopback: True, adds loopback structures.
        straight_separation: from edge to edge.
        list_port_labels: None, adds TM labels to port indices in this list.
        connected_port_list_ids: names of ports only for type 0 optical routing.
        nb_optical_ports_lines: number of grating coupler lines.
        force_manhattan: False
        excluded_ports: list of port names to exclude when adding gratings.
        grating_indices: list of grating coupler indices.
        routing_straight: function to route.
        routing_method: get_route.
        optical_routing_type: None: auto, 0: no extension, 1: standard, 2: check.
        gc_rotation: fiber coupler rotation in degrees. Defaults to -90.
        input_port_indexes: to connect.

    """
    c = gf.Component()
    component = gf.get_component(component)
    r = c << gf.routing.add_fiber_array(
        component=component,
        grating_coupler=grating_coupler,
        with_loopback=with_loopback,
        fiber_spacing=fiber_spacing,
        layer_label=None,
        **kwargs,
    )
    optical_ports = r.get_ports_list(port_type="optical")
    c.add_ports(optical_ports)

    electrical_ports = r.get_ports_list(
        port_type="electrical", orientation=electrical_port_orientation
    )
    electrical_port_names = electrical_port_names or [p.name for p in electrical_ports]

    npads = npads or len(electrical_port_names)
    pads = c << gf.components.array(
        component=pad,
        columns=npads,
        spacing=(pad_spacing, 0),
    )
    pads.x = r.x
    pads.ymin = r.ymin + pad_gc_spacing

    electrical_ports = [r[por_name] for por_name in electrical_port_names]
    nroutes = min(len(electrical_ports), npads)

    ports1 = electrical_ports[:nroutes]
    ports2 = pads.get_ports_list(orientation=270)[:nroutes]
    routes = gf.routing.get_bundle_electrical(
        ports1=ports1,
        ports2=ports2,
        cross_section=xs_metal,
        enforce_port_ordering=False,
    )
    for route in routes:
        c.add(route.references)

    c.add_ports(ports2)
    xc, yc = getattr(r.size_info, anchor)

    analysis_settings = analysis_settings or {}
    analysis_settings.update(component.metadata.get("full", {}))

    if layer_label:
        settings = dict(
            name=component.name,
            measurement=measurement,
            xopt=[int(optical_ports[0].x - xc)],
            yopt=[int(optical_ports[0].y - yc)],
            xelec=[int(ports2[0].x - xc)],
            yelec=[int(ports2[0].y - yc)],
            measurement_settings=measurement_settings,
            analysis=analysis,
            analysis_settings=analysis_settings,
            doe=doe,
        )
        info = json.dumps(settings)
        c.add_label(layer=layer_label, text=info, position=(xc, yc))

    c.copy_child_info(r)
    return c


if __name__ == "__main__":
    gf.config.rich_output()
    c = add_fiber_array_optical_south_electrical_north(
        measurement_settings={"wavelength_min": 1550}
    )

    d = json.loads(c.labels[0].text)
    print(d)
    # import gdsfactory as gf
    # from functools import partial

    # component = partial(mzi_phase_shifter, length_y=1)
    # c = add_fiber_array_optical_south_electrical_north(
    #     component=component,
    #     electrical_port_names=["top_l_e2", "top_r_e2"],
    #     npads=5,
    # )
    # component = partial(gf.c.ring_single_heater, length_x=10)
    # c = add_fiber_array_optical_south_electrical_north(
    #     component=component,
    #     electrical_port_names=["l_e2", "r_e2"],
    #     npads=5,
    # )
    c.show(show_ports=False)
